PluginInterface.require_command("ogginfo")
from mx import DateTime

class OggVorbisFile(PluginInterface.File):

	date = None
	sample_rate = None
	track_number = None
	bitrate = None
	title = None
	album = None
	genre = None
	artist = None
	# fixme implement comment and length extraction
# 	comment = None
# 	length = None

	OggVorbisFile_catalogable = [
		"date","sample_rate","track_number","bitrate",
		"title","album","genre","artist","comment","length" ]

	def __init__(self,filename,mimetype,volume):
		PluginInterface.File.__init__(self,filename,mimetype,volume)
		metadata = ogginfo_extract(filename)
		for key,value in metadata.iteritems():
			if key == "date":
				self.date = DateTime.DateTime(value,1,1)
			elif key == "Rate":
				self.sample_rate = value
			elif key == "tracknumber":
				self.track_number = value
			elif key == "title":
				self.title = value
			elif key == "album":
				self.album = value
			elif key == "Average bitrate":
				self.bitrate = value
			elif key == "genre":
				self.genre = value
# 			elif key == "id3.comment" and len(value) > 0:
# 				self.comment = value
# 			elif key == "technical.length":
# 				self.length = value
			elif key == "artist":
				self.artist = value

	def __repr__(self):
		return "Ogg Vorbis music file %s (%s - %s - %s)"%(self.get_filename(),self.artist,self.album,self.title)

def get_factory():
	return OggVorbisFile

def get_mimetypes():
	return ["audio/vorbis","application/ogg"]


def ogginfo_extract(filename):
	import commands
	output = commands.getoutput("LANG=en_US.UTF-8 ogginfo %s"%PluginInterface.shell_quote(filename))
	output = PluginInterface.transcode_to_platform_encoding(output,"UTF-8")
	lines = output.splitlines()
	lines = [ a.strip() for a in lines ]
	tags = [ a.split("=",1) for a in lines if a.find("=") != -1 ]
	techinfo = [ a.split(":",1) for a in lines if a.find(":") != -1 ]
	t = {}
	for tag in tags:
		key,value = tag[0],tag[1]
		key = key.strip()
		value = value.strip()
		t[key] = value
	for techin in techinfo:
		key,value = techin[0],techin[1]
		key = key.strip()
		value = value.strip()
		t[key] = value
	
	key = "date"
	if t.has_key(key):
		try: t[key] = int(t[key])
		except:
			PluginInterface.log("vorbis-tools-audio-vorbis").warning("Error decoding year (%s)",t[key],exc_info=True)
			del(t[key])
	key = "tracknumber"
	if t.has_key(key):
		try: t[key] = int(t[key])
		except:
			PluginInterface.log("vorbis-tools-audio-vorbis").warning("Error decoding track (%s)",t[key],exc_info=True)
			del(t[key])
# 	key = "COMM"
# 	if t.has_key(key):
# 		if t[key].startswith("(ID3v1 Comment)["): t[key] = t[key][22:]
	key = "Average bitrate"
	if t.has_key(key):
		try: t[key] = float(t[key].split(" ")[0])*1000
		except: 
			PluginInterface.log("vorbis-tools-audio-vorbis").warning("Error decoding bitrate (%s)",t[key],exc_info=True)
			del(t[key])
	key = "Rate"
	if t.has_key(key):
		try:
			t[key] = int(t[key])
		except: 
			PluginInterface.log("vorbis-tools-audio-vorbis").warning("Error decoding frequency (%s)",t[key],exc_info=True)
			del(t[key])
	
	return t



from Component import Component
from ControllableThread import ControllableThread


class Service(Component,ControllableThread):
	"""The base class for all service processes."""
	def __init__(self,public_interface=None):
		Component.__init__(self,public_interface=public_interface)
		ControllableThread.__init__(self)



from MetadataService.InterfaceServer import InterfaceServer
from SocketXMLRPCServer import SocketXMLRPCServer
import os
import ConfigParser
from MetadataService.common import from_xml_params,to_xml_params
from SimpleXMLRPCServer import SimpleXMLRPCDispatcher

class SocketXMLRPCInterfaceServer(InterfaceServer,SocketXMLRPCServer):
	""" A self-documenting implementation of the public interface server using standard TCP/HTTP transport """
	
	def __init__(self,config,interfaces):
		InterfaceServer.__init__(self,interfaces=interfaces,config=config)
		self.set_component_name("InterfaceServer.SocketXMLRPC")

		try: sockaddr = self.config.get("socket interface","socket")
		except (ConfigParser.NoSectionError,ConfigParser.NoOptionError):
			sockaddr = os.path.join(self.config.get_data_dir(),"socket")
		try: os.unlink(sockaddr)
		except OSError,e:
			if e.errno == 2: pass
			else: raise
		SocketXMLRPCServer.__init__(self,sockaddr)
		self.register_instance(self.public_interfaces)
		self.register_introspection_functions()
		self.register_multicall_functions()
		self.setDaemon(True) # fixme, this avoids the blocking upon shutdown
		
	def run(self):
		self.serve_forever()

	# FIXME: this is the place to determine user ID and submit it to the public interface
	def _dispatch(self,method,params):
		params = from_xml_params(params)
		return to_xml_params(SimpleXMLRPCDispatcher._dispatch(self,method,params))



############################################################################
#    Copyright (C) 2005 by Manuel Amador (Rudd-O)                          #
#    rudd-o@amautacorp.com                                                 #
#                                                                          #
#    This program is free software; you can redistribute it and#or modify  #
#    it under the terms of the GNU General Public License as published by  #
#    the Free Software Foundation; either version 2 of the License, or     #
#    (at your option) any later version.                                   #
#                                                                          #
#    This program is distributed in the hope that it will be useful,       #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of        #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         #
#    GNU General Public License for more details.                          #
#                                                                          #
#    You should have received a copy of the GNU General Public License     #
#    along with this program; if not, write to the                         #
#    Free Software Foundation, Inc.,                                       #
#    59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             #
############################################################################

# FIXME implement an interface that allows the indexer to compare in-store file objects' interface versions with its corresponding plugin"s interface version, to provide for automatic schema migration.  Bump the interface number on your plugin, and pof all files are automatically reindexed when they are modified (or perhaps upon startup/reload)
# FIXME implement an observer pattern so that the index database detects when the plugins are reloaded, suspend database access to client classes, run __prepare_indexes and refresh objects whose interfaces have changed (ay que pensar bien el orden de mierda de estas operaciones para que no se vaya todo a la verga"), 

import sys
import os
import glob
import imp
import urllib

from gnome import vfs

# FIXME PRETTY PLEASE
# this may involve the creation of two other data types: naivedate (yy/mm/dd) and naivetime(hh:mm:ss) which operate in local time, and then client methods can interpret them and cast them to DateTime on their own, with this knowledge.
# or perhaps even better create an object daydelta = num of seconds after 0h00 and then we can sum that to the DateTime
# this should be moved to KDE's own module
def get_mimetype(filename):
		""" Returns a file's MIME type. """
		if os.path.isdir(filename): return "inode/directory"
		fn = urllib.pathname2url(filename)
		try: mt = vfs.get_mime_type(fn)
		except RuntimeError: raise IOError,(2,"No such file or directory: %s"%repr(filename))
		if len(mt) > 0: return mt




############################################################################
#    Copyright (C) 2005 by Manuel Amador (Rudd-O)                          #
#    rudd-o@amautacorp.com                                                 #
#                                                                          #
#    This program is free software; you can redistribute it and#or modify  #
#    it under the terms of the GNU General Public License as published by  #
#    the Free Software Foundation; either version 2 of the License, or     #
#    (at your option) any later version.                                   #
#                                                                          #
#    This program is distributed in the hope that it will be useful,       #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of        #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         #
#    GNU General Public License for more details.                          #
#                                                                          #
#    You should have received a copy of the GNU General Public License     #
#    along with this program; if not, write to the                         #
#    Free Software Foundation, Inc.,                                       #
#    59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             #
############################################################################

# FIXME implement an interface that allows the indexer to compare in-store file objects' interface versions with its corresponding plugin"s interface version, to provide for automatic schema migration.  Bump the interface number on your plugin, and pof all files are automatically reindexed when they are modified (or perhaps upon startup/reload)
# FIXME implement an observer pattern so that the index database detects when the plugins are reloaded, suspend database access to client classes, run __prepare_indexes and refresh objects whose interfaces have changed (ay que pensar bien el orden de mierda de estas operaciones para que no se vaya todo a la verga"), 

import os
from MetadataService.Component import Component
import platform

import PluginInterface
import __builtin__
__builtin__.PluginInterface = PluginInterface

class FilePluginManager(Component):

	mimetype_plugin_map = {}
	log = None

	def __init__(self):
		Component.__init__(self,public_interface=None)
		self.set_component_name("FilesystemAgent.FilePluginManager")
		self.reload_plugins()

	def reload_plugins(self):
		#Fixme: need to implement observer pattern so index database can rebuild indexes as necessary, or perhaps the indexer determines what indexes to build in a different way, seeing that there are different infospaces to build into this service in the medium-term future
		
		self._new_mimetype_plugin_map = {}
		self.load_plugins()
		self.mimetype_plugin_map = self._new_mimetype_plugin_map
# 		self.log.debug("Mimetypes registered: %s",str(self.mimetype_plugin_map))
	

	def load_plugin(self,plugin):
		if plugin == "__init__.py": return # do not load __init__.py!
		import imp
		module_name = plugin[0:len(plugin)-len(".py")]
		
		try: fd, filename, module_info = imp.find_module(module_name,[self.get_plugin_path()])
		except ImportError: raise IOError,(13,"Permission denied: '%s'"%a)
		
		module = imp.load_module(module_name,fd,filename,module_info)
		fd.close()
		
		if module.__dict__.has_key("get_mimetypes") and module.__dict__.has_key("get_factory"):
			# only then we'll consider the module as a plugin
			for mimetype in module.get_mimetypes():
				self._new_mimetype_plugin_map[mimetype] = module.get_factory()
			self.log.debug("Loaded plugin ("+plugin+"), handling " + str(module.get_mimetypes()))
		else: self.log.warning("Not a plugin: " + plugin)

	def get_plugin_path(self):
		import FilePluginManager
		return os.path.join(os.path.dirname(FilePluginManager.__file__),"plugins")
		
	def load_plugins(self):
		self.log.debug("Loading indexer plugins")
		olddir = os.getcwd()
		os.chdir(self.get_plugin_path())
		pluginextension = ".py"
		from glob import glob
		pythonpluginfiles = glob("*"+pluginextension)
		os.chdir(olddir)
		for a in pythonpluginfiles:
			try: self.load_plugin(a)
			except ImportError, e:
				self.log.warning("Plugin %s requires unavailable libraries or is unreadable",a,exc_info=False) # we need a global function or construct which can let us do exc_info = IfDebug, and a debugging config switch
			except IOError, e: self.log.warning("Plugin %s cannot be opened for reading",a,exc_info=True)
			except: self.log.warning("Cannot load %s",a,exc_info=True)
		self.log.info("Indexer plugins loaded")

	def get_factory_for_mimetype(self,mimetype):
		assert type(mimetype) is str and len(mimetype) > 0
		if self.mimetype_plugin_map.has_key(mimetype): return self.mimetype_plugin_map[mimetype]
		return None
	
	def get_information_object(self,filename,volume):
		from InfospaceObjects import Volume
		assert isinstance(volume,Volume)
		assert type(filename) is str and len(filename) > 0
		mimetype = platform.get_mimetype(filename)
		if mimetype is None: mimetype = "application/octet-stream" # just in case no mimetype is detected!
		factory = self.get_factory_for_mimetype(mimetype)
		if factory is None: #we'll default to builtins
			if os.path.isdir(filename):
				from Directory import Directory
				factory = Directory
			else:
				from File import File
				factory = File
		object = factory(filename=filename,volume=volume,mimetype=mimetype)
		return object



#!/usr/bin/env python

from threading import Thread
import logging

class ControllableThread(Thread):
	"""The ControllableThread is a subclass of Thread that provides new methods kill() and stop().  In your run() method, you should periodically check if the thread has been killed with self.killed(), which will return True if another thread process your thread.  You can implement the stop() method, which will be called when your thread is killed(). Example implementation:

	def run(self):
		while not self.killed():
			... do some stuff that should not take long

	def stop(self):
		...do some cleanups
	"""
	__killed = False

	def __init__(self):
		Thread.__init__(self)

	def killed(self): return self.__killed

	def kill(self,asynchronous=False):
		self.__killed = True
		if self.isAlive():
			self.on_kill()
			if asynchronous is False and not self.isDaemon(): # only join non-daemon threads
				self.join()

	def on_kill(self):
		"""Implement this method to have something performed when the thread is explicitly killed.  This will not be performed if the thread aborts due to an Exception or terminates its run() method on its own."""
		pass


class VolumeRepresentation:
	filesystem_type = None
	guid = None
	label = None
	current_device_file = None
	removable = False
	loop_mounted = False
	network_volume = False

	def __str__(self):
		return "volume representation, label %s, current device %s, GUID %s, filesystem type %s, removable %s"%(self.label,self.current_device_file,self.guid,self.filesystem_type,self.removable)
		
class MtabRecord:
	device_file = None
	mountpoint = None
	filesystem_type = None
	options = None
	col5 = None
	col6 = None
	
	def __init__(self,device_file,mountpoint,filesystem_type,options,col5,col6):
		if device_file != "none": self.device_file = device_file
		self.mountpoint = mountpoint
		self.filesystem_type = filesystem_type
		self.options = options
		self.col5 = col5
		self.col6 = col6
		
	def __str__(self):
		return "'%s' '%s' '%s' '%s' '%s' '%s'"%(self.device_file,
				self.mountpoint,self.filesystem_type,self.options,self.col5,self.col6)
	

class Mtab:
	records = []
	def add_record(self,device_file,mountpoint,filesystem_type,options,col5,col6):
		self.records.append(MtabRecord(device_file,mountpoint,filesystem_type,options,col5,col6))
		
	def __str__(self): return "mtab composed of records: %s"%[str(r) for r in self.records]
	



from Queue import Queue, Full, Empty
Full=Full
Empty=Empty

class Producer:
	"""A producer is a process which produces elements and appends them to an internal queue.  Producer processes (generally the run() method in threads - see below) should append elements to their queues by using self.produce(element,max_wait=0).
	  
	The queue has a finite length, and thus appending can block indefinitely (if max_wait is unspecified) or fail with a Full exception after max_wait seconds.  The producer's queue can be emptied by calling empty_queue().  Clients of the Producer class (generally Consumers) can get the next element in the list with get_next_element(max_wait=0) which can fail with an Empty exception.
	
	Producers are designed to be mixed-in with Thread classes.  You need to initialize Producers specifying the queue's maximum length.
	"""
	
	def __init__(self,max_length):
		self.__queue = Queue(max_length)
		
	def produce(self,element,max_wait=0):
		if max_wait == 0: self.__queue.put_nowait(element)
		else: self.__queue.put(element,True,max_wait)
	
	def empty_queue(self):
		try:
			while True: dummy = self.queue.get_nowait()
		except Empty: pass

	def get_next_element(self,max_wait): # for use of the Consumer class
		if max_wait == 0: return self.__queue.get_nowait()
		else: return self.__queue.get(True,max_wait)
		
	
# def abort_consumption() using events instead of locking queues?
class Consumer:
	"""A Consumer is a process which consumes Producer queue elements, one at a time.  Consumer processes can be initialized with or without a producer, but need to be told sooner or later which Producer to consume, with consumer_instance.consume_from(producer_instance).  To consume an element in the process (generally the run() method in threads - see below), use the self.consume(max_wait) method, which can raise the Empty exception if there is no element to consume, or no producer to consume from.
	
	Consumers are designed to be mixed-in with Thread classes.  You can either initialize Consumers specifying the producer instance you want to consume from, or initialize them without a producer instance, and later tell consumers which producer to consume from, with consumer_instance.consume_from().
	"""
	
	__producer = None
	
	def __init__(self,producer=None):
		if producer is None: pass
		else: self.consume_from(producer)
			
	def consume_from(self,producer):
		assert isinstance(producer,Producer)
		self.__producer = producer
	
	def consume(self,max_wait=0):
		"""Pops an element from the current producer's queue.  Waits at most max_wait seconds for the queue to have an element before raising Empty"""
		try: a = self.__producer.get_next_element(max_wait)
		except AttributeError:
			time.sleep(max_wait)
			raise Empty
		return a



#!/usr/bin/env python

#FIXME make a check that checks after 2 seconds of having actually started services up, to see if they are still up, and if they are not, shutdown the whole thing safely

import signal
import os

import logging

#FIXME make these imports dynamic, perhaps by making the components themselves load dynamically on request as a plugin fashion
# Machinery imports
from Index import Index
from Databases import ZODBDatabase
from pgtools.Catalog import PGCatalog
from pgtools.DatabaseCluster import PGDatabaseCluster
from AdaptiveLoadController import AdaptiveLoadController
from InfospaceAgents.Filesystem.FilesystemAgent import FilesystemAgent
from InterfaceServers.SocketXMLRPC.SocketXMLRPCInterfaceServer import SocketXMLRPCInterfaceServer
from InterfaceServers.XMLRPC.XMLRPCInterfaceServer import XMLRPCInterfaceServer
from Component import Component
from PublicInterface import PublicInterface
import ConfigParser
import Config
import stat

class TerminateRequest(Exception):
	"""A dummy class to use whenever we catch a signal"""

# startup process: setup logging, initialize index(with database), initialize load controller, initialize infospace agents(index) and register them with the index, initialize interface servers ([load controller] + [index] + infospace_agents), start them in that order
# shutdown process: stop interface servers, stop infospace agents, stop load controller, close index, end logging.
class MasterService(Component):
	config = None
	index = None
	dbcluster = None
	load_controller = None
	infospace_agents = []
	interface_servers = []
# 	log = None
	
	def __init__(self,config_filename=None):
		Component.__init__(self,public_interface=MetadataServicePublicInterface("metadataService",self))
		self.set_component_name("MetadataService")
		self.config = Config.get(config_filename)

	def setup_logging(self):
		log = logging.getLogger()
		log.setLevel(self.config.getint("logging","log level"))

		stderr_handler = logging.StreamHandler()
		stderr_format = "%(message)s"
		stderr_formatter = logging.Formatter(stderr_format)
		stderr_handler.setFormatter(stderr_formatter)
		log.addHandler(stderr_handler)

		logfile = None
		try: logfile = self.config.get("logging","log file")
		except ConfigParser.NoOptionError: pass

		if logfile:
			format = "%(levelname)s:%(name)s: %(message)s"
			formatter = logging.Formatter(format)
			if logfile == "stderr":
				handler = stderr_handler
				handler.setFormatter(formatter)
			else:
				handler = logging.FileHandler(logfile)
				handler.setFormatter(formatter)
				log.addHandler(handler)


		# if the user didn't configure a log file, or the log file is not stderr
		# we just set the default stderr handler to ERROR
		if not logfile or logfile != "stderr":
			stderr_handler.setLevel(logging.ERROR)

		syslog = self.config.getboolean("logging","use syslog")
		if syslog:
			facility = self.config.getint("logging","log facility")
			handler = logging.handlers.SysLogHandler(address="/dev/log",facility=facility)
			format = "%(name)s: %(message)s"
			formatter = logging.Formatter(format)
			handler.setFormatter(formatter)
			log.addHandler(handler)

		log.debug("Using config file %s",self.config.get_filename())
		
	def install_sighandlers(self):
		signal.signal(signal.SIGINT,self.__signal_handler)
		signal.signal(signal.SIGTERM,self.__signal_handler)
		signal.signal(signal.SIGHUP,self.__signal_handler)
	
	def end_logging(self):
		logging.shutdown()
	
	def initialize_dbcluster(self):
		indexpath = self.config.get("index","path")
		if not os.path.isdir(indexpath):
			os.makedirs(indexpath)
			os.chmod(indexpath,stat.S_IRUSR|stat.S_IWUSR|stat.S_IXUSR)
		self.log.debug("Starting database cluster")
		self.dbcluster = PGDatabaseCluster(os.path.join(indexpath,"catalog"))
		self.dbcluster.start()
		self.log.debug("Database cluster started")
	
	def initialize_index(self):
		# fixme make modular
		assert self.dbcluster is not None
		
		indexpath = self.config.get("index","path")
		self.log.debug("Starting index")
		database = ZODBDatabase(os.path.join(indexpath,"database"))
		catalog_arguments = self.dbcluster.get_database("catalog")
		catalog = PGCatalog(**catalog_arguments)
		self.index = Index(database=database,catalog=catalog)
		self.log.debug("Index started")
	
	def close_index(self):
		if self.index:
			self.log.debug("Closing index")
			self.index.close()
			self.log.debug("Index closed")
			
	def close_dbcluster(self):
		if self.dbcluster:
			self.log.debug("Stopping database cluster")
			try:
				self.dbcluster.stop_if_started()
				self.log.debug("Database cluster stopped")
			except: self.log.error("Could not stop database cluster",exc_info=True)
	
	def initialize_load_controller(self):
		load = self.config.getfloat("load controller","target load")
		self.load_controller = AdaptiveLoadController(target_load=load)
	
	def initialize_infospace_agents(self):
		# fixme make modular
		assert isinstance(self.index,Index)
		i = FilesystemAgent(index=self.index,config=self.config)
		self.infospace_agents.append(i)
		self.index.register_infospace_agent(i)
	
	def initialize_interface_servers(self):
		# fixme make modular
		components = self.get_components()
		assert len(components) > 0 #fixme make modular and organize interface servers folder into reasonable packages
		
		interfaces = [ c.public_interface for c in self.get_components() if c.public_interface is not None ]
		
		self.interface_servers.append(SocketXMLRPCInterfaceServer(interfaces=interfaces,config=self.config))
		
		self.interface_servers.append(XMLRPCInterfaceServer(interfaces=interfaces,config=self.config))
	
	def get_services(self): return [self.load_controller] + self.infospace_agents + self.interface_servers

	def get_components(self): return [self,self.index] + self.get_services()

	def start_services(self):
		self.log.debug("Starting services")
		services = self.get_services()
		
		try:
			for service in services:
				self.log.debug("Starting service %s",service.get_component_name())
				service.start()
				self.log.debug("Service %s started",service.get_component_name())
		except TerminateRequest: raise # This will be caught in the run() method
		except:
			self.log.critical("Failed to start service %s, startup aborted",service.get_component_name(),exc_info=True)
			raise
		
		self.log.debug("Services started")
	
	def stop_services(self):
		self.log.debug("Stopping services")
		
		services = self.get_services()
		services.remove(self.load_controller)
		services.append(self.load_controller)
		services.reverse()
		for service in services:
			try:
				self.log.debug("Stopping %s",service.get_component_name())
				if not service.isAlive():
					self.log.error("Service %s had died prematurely",service.get_component_name())
				else:
					service.kill()
					self.log.debug("Service %s stopped",service.get_component_name())
			except AttributeError: pass
			except:
				self.log.error("Cannot stop service %s",service,exc_info=True)

		self.log.debug("Services stopped")

	def run(self):
		# the setup phase
		self.setup_logging()
		self.install_sighandlers()
		self.log.info("Metadata service starting")
		retvalue = 0
		
		# the start phase
		errmsg = None
		try:
			self.initialize_dbcluster()
			try: self.initialize_index()
			except (IOError,OSError), e:
				errmsg = "Metadata service cannot access the database directory %s.  Please check your configuration and the directory permissions."%self.config.get("index","path")
				raise
			self.initialize_load_controller()
			self.initialize_infospace_agents()
			self.initialize_interface_servers()
			self.start_services()
		except TerminateRequest,signum:
			self.log.warning("Metadata service interrupted by signal %s while starting sub-services",signum)
			retvalue = 8
		except:
			if errmsg: self.log.critical(errmsg,exc_info=True)
			else: self.log.critical("Metadata service failed to start one of the sub-services, cannot continue",exc_info=True)
			retvalue = 4 # an error initializing sub-services


		
		# the run phase
		if retvalue == 0:
			self.log.info("Metadata service started")
			try:
				while True: signal.pause() # this should rather wait for events on a queue instead of staying here with signal.pause() -  events like TerminateRequest, ReloadConfigRequest, and, of course, the signal handlers (and the public interface for the metadata service!) should feed the queues with events.
			except TerminateRequest,sig: self.log.info("Metadata service ending on signal %s",sig)
			except:
				self.log.error("Anomalous exception in signal.pause()",exc_info=True)
				retvalue = 16
		
		# the stop phase
		self.stop_services()
		self.close_index()
		self.close_dbcluster()
		self.log.info("Metadata service ended")
		self.end_logging()
		return retvalue
	
	def request_exit(self,signum=signal.SIGTERM): # these actually do stuff
		self.__block_sighandlers() # this to avoid signals to the process while exiting
		raise TerminateRequest, signum
	
# 	def reload_config(self): # these actually do stuff
# 		self.block_sighandlers() # this to avoid signals to the process while exiting
# 		raise NotImplemented # implement!
# 		self.unblock_sighandlers() # restore previous condition
	
	__signal_block = False
	def __block_sighandlers(self): self.__signal_block = True
	def __unblock_sighandlers(self): self.__signal_block = False
	
	def __signal_handler(self,signum,arg2):
		if self.__signal_block: return
# 		if signum == signal.SIGHUP: self.reload_config()
		if signum in (signal.SIGTERM, signal.SIGINT): self.request_exit(signum)

# 	def request_exit(self): os.kill(os.getpid(),signal.SIGTERM) # we need to do it thsi way because if we raise a terminate request it will be raised in the context of the calling process.
# 	def request_reload_config(self): os.kill(os.getpid(),signal.SIGHUP)  #ditto

#Write the interface for this metadata service, and migrate to Component semantics


class MetadataServicePublicInterface(PublicInterface):
	def shutDown(self):
		os.kill(os.getpid(),signal.SIGTERM)
		return True



class PlainTextFile(PluginInterface.File):

# 	format = None
# 	lines = None
	characters = None
	contents = None
	encoding = None
	
	PlainTextFile_catalogable = [ "characters","contents","encoding" ]


	def __init__(self,filename,mimetype,volume):
		PluginInterface.File.__init__(self,filename,mimetype,volume)
# 		metadata = get_kde_metadata_expert().get_metadata(filename)
# 		if metadata.has_key("general.format"):
# 			self.format = metadata["general.format"]
# 		if metadata.has_key("general.lines"):
# 			self.lines = int(metadata["general.lines"])
# 		if metadata.has_key("general.characters"):
# 			self.characters = int(metadata["general.characters"])
		# now we'll read the file contents
		f = file(filename,"rb")
		c = f.read(-1)
		encoding = PluginInterface.detect_encoding(c)
		f.close()
		try:
			if encoding:
				self.encoding = encoding
				c = PluginInterface.transcode_to_platform_encoding(c,encoding)
				self.contents = PluginInterface.remove_whitespace(c)
				for char in self.contents:
					if ord(char) < 32 and char not in ['\t','\n','\r']:
						# these are control characters, inadmissible in text files
						# so we set the contents to None and then go on
						self.contents = None
						break
		except (UnicodeDecodeError,UnicodeEncodeError):
			PluginInterface.log("text-plain").info("%s is in an encoding unsupported by the platform",filename,exc_info=True)
		# FIXME we're just assuming text files are in platform encoding, and that is EVIL
		self.characters = len(c)

	def __str__(self):
		return "plain text file %s (%d characters)"%(self.get_filename(),self.characters)


def get_factory():
	return PlainTextFile

def get_mimetypes():
	return ["text/plain"]



import linux
import GNOME

try: get_mimetype = GNOME.get_mimetype
except ImportError: get_mimetype = linux.get_mimetype
get_currently_known_volumes = linux.get_currently_known_volumes
get_mtab = linux.get_mtab
is_executable = linux.is_executable
strip_mountpoint = linux.strip_mountpoint
is_or_is_below = linux.is_or_is_below


import os
import mimetypes
# FIXME these are mostly stubs, no more

def get_mimetype(filename):
	if os.path.isdir(filename): return "inode/directory"
	else:
		tupl = mimetypes.guess_type(filename)
		mimetype = tupl[0]
		if mimetype is None: return "application/octet-stream"
		return mimetype


# fixme this should receive a queue instead of an indexer?
def get_indexer(index=None):
	if not get_indexer.__dict__.has_key("active"):
		if index is None: raise NotInitializedException
		get_indexer.active = FilesystemIndexer(index)
	return get_indexer.active


"""This is a proxy module for common functions used by file plugins"""

from MetadataService import common
from File import File


# FIXME: Document that this is the official plugin utilities interface
is_utf8_encoded = common.is_utf8_encoded
transcode_to_platform_encoding = common.transcode_to_platform_encoding
platform_to_utf8 = common.platform_to_utf8
utf8_to_platform = common.utf8_to_platform
transcode_to_utf8_encoding = common.transcode_to_utf8_encoding
transcode = common.transcode
local_timestamp_to_utc = common.local_timestamp_to_utc
utc_timestamp_to_local = common.utc_timestamp_to_local
shell_quote = common.shell_quote
remove_whitespace = common.remove_whitespace
detect_encoding = common.detect_encoding
File = File
from platform import is_executable
def require_command(command):
	if type(command) is not str or len(command) < 1:
		raise ValueError, "supplied command is invalid"
	if not is_executable(command): raise ImportError, "Plugin requires unavailable %s command" % repr(command)
		
def require_commands(commands):
	if type(commands) is not list: raise TypeError, "require_commands needs a list of commands"
	for command in commands: require_command(command)

import platform
platform = platform

import logging
def log(channel):
	assert type(channel) is str and len(channel) > 0
	return logging.getLogger("Plugins.%s"%channel)

extractors = {}
try: import libextractor; extractors["libextractor"] = libextractor.get_metadata
except: pass

def get_extractor(name):
	assert type(name) is str and len(name) > 0
	try: return extractors[name]
	except KeyError: raise ImportError, "%s is not available on this system"%name



import time
from common import DateTime,platform_to_utf8
from xmlrpclib import DateTime as xmlDateTime
import ZODB
from persistent import Persistent
from pgtools.Catalog import Catalogable

### make it compliant with the dublin core FIXME, or maybe dublin core compliance must be in File

class InformationObject(Persistent,Catalogable):
	"""This is the base class for all information objects stored in the index.  Indexer implementors need to subclass from this, and implement all methods, in order to be able to store their information objects safely into the index.  Do not forget to call InformationObject.__init__(...) in your inherited class."""
	
	indexing_date = None
	
	InformationObject_catalogable = "indexing_date"
	
	def __init__(self):
		self.indexing_date = DateTime.gmtime(time.time())
		
# 	def is_valid(self):  # fixme think this through
		"""You should use this method to let the index determine whether the file is valid.  Index will automatically purge all objects which are not valid."""
		
	# maybe FIXME the responsibility for this conversion relies in the XML-RPC interface
	def as_marshallable_representation(self):
		""" Returns an XML-serializable representation of the object in a dictionary.  Strings are converted to UTF-8 FIXME prove that's needed, and dates/times are in the DateTime wrapper object used by xmlrpclib. """
		element = {}
		for key,value in self.__dict__.iteritems():
 			if type(value) is str: value = platform_to_utf8(value) # utf8 needed?
			elif type(value) is DateTime.DateTimeType:
				# this should return an xmlDateTime object representing the time in UTC (override a stupid mistake in DateTime from the xmlrpclib
				value = xmlDateTime(value.localtime()) # FIXME verify this is working, on-the-wire!
			elif type(value) in (float,int,long): value = value
			else: value = str(value)
			element[key] = value # FIXME need to process other types
		return element
		
	def __str__(self):
		return "information object"



#!/usr/bin/env python

import SocketServer
from SimpleXMLRPCServer import SimpleXMLRPCDispatcher , SimpleXMLRPCRequestHandler

class SocketXMLRPCRequestHandler(SimpleXMLRPCRequestHandler):
	def log_request(self, code='-', size='-'):
		"""Log an accepted request.  Parent class method vomits on UNIX sockets. """
# 		logging.getLogger("interfaceserver").debug("Acepted request: %s", self.requestline)
	
	def log_error(self, *args):
		"""Log an error.  Parent class method vomits on UNIX sockets. """
# 		logging.getLogger("interfaceserver").error("Request error: %s", str(args))

class SocketXMLRPCServer(SocketServer.UnixStreamServer, SimpleXMLRPCDispatcher):

	def __init__(self,socketaddr,requestHandler=SocketXMLRPCRequestHandler,logRequests=1):
		self.logRequests = logRequests
		SimpleXMLRPCDispatcher.__init__(self)
		SocketServer.UnixStreamServer.__init__(self,socketaddr,requestHandler)



import os
import commands
from PlatformObjects import VolumeRepresentation, Mtab
import re
# FIXME these are mostly stubs, no more

_blkid_datum_regexp = re.compile('(\w+)="(.+?)"')
_network_fstypes = "nfs","smbfs","autofs","autofs4","ncpfs","smb","coda","afs"

def is_executable(command):
	if os.access(command,os.X_OK): return True
	path = os.environ["PATH"].split(":")
	for direct in path:
		if os.access(os.path.join(direct,command),os.X_OK): return True
	return False

if not is_executable("/sbin/blkid"): raise ImportError, "Required command /sbin/blkid is not available"

def get_currently_known_volumes():
	""" returns a dictionary of volumes, keyed by GUID.  Each volume is a struct VolumeRepresentation, with GUID, currently known device file, filesystem type, and volume label information."""
	output = commands.getoutput("/sbin/blkid")
	volumes = {}
	for line in output.splitlines():
		dline = line.split(':',2)
		datums = _blkid_datum_regexp.findall(dline[1].strip())
		data = VolumeRepresentation()
		data.current_device_file = dline[0]
		data.removable = False # STUB!!!!! need to be fixed for cdroms and shit FIXME
		data.loop_mounted = data.current_device_file.startswith("/dev/loop")
		for datum in datums:
			key = datum[0]; value = datum[1]
			if key == "TYPE":
				data.filesystem_type = value
				if data.filesystem_type in _network_fstypes:
					data.network_volume = True
			if key == "UUID": data.guid = value
			if key == "LABEL": data.label = value
		if data.guid:	# this covers the famous devfs, swap, procfs and rpc_pipefs,
					# sysfs cases in a very handsome fashion, yippeeeee
			volumes[data.guid] = data
	return volumes


def get_mtab():
	""" return the mtab as a list of lists """
# 	import csv
# 	
# 	class MtabDialect(csv.Dialect):
# 	
# 		delimiter = " "
# 		lineterminator = "\n"
# 		escapechar = '\\'
# 		skipinitialspace = True
# 		doublequote = True
# 		quotechar = ' '
# 		quoting = csv.QUOTE_NONE
	
	f = file("/etc/mtab","r")
	mtablines = f.readlines()
	f.close()
	
	mtablines = [ line.split(" ") for line in mtablines ]
	
	mtab = Mtab()
	
	for a in mtablines:
		mtab.add_record(a[0].decode("string_escape"),a[1].decode("string_escape"),
						a[2],a[3],a[4],a[5])
	return mtab


def strip_mountpoint(filename,mountpoint):
	filename = os.path.normpath(filename)
	mountpoint = os.path.normpath(mountpoint)
# 	if not mountpoint.startswith(mountpoint): raise RuntimeError, "%s not in %s"%(filename,mountpoint)
	while mountpoint.endswith("/"): mountpoint = mountpoint[:-1]
	filename =  filename[len(mountpoint)+1:]
	return filename

def _canonicalize_path(path):
	path = os.path.normpath(path)
# 	while path.endswith("/"): path = path[:-1]
	if not path.endswith("/"): path = path + "/"
	return path

def is_or_is_below(filename,path):
	""" Tests whether filename is the same as path or is below path """
	filename = _canonicalize_path(filename)
	path = _canonicalize_path(path)
	return filename.startswith(path)

from MetadataService.common import shell_quote
def get_mimetype(filename):
	if os.path.isdir(filename): return "inode/directory"
	output = commands.getoutput("file -bi %s"%shell_quote(filename))
	if output.startswith("cannot"): return None
	mimetype = output.split(";",2)[0].strip()
	if len(mimetype) > 0: return mimetype
	
	



#!/usr/bin/env python

from mx import DateTime
from mx.DateTime import DateTimeType
DateTimeType = DateTimeType
DateTime = DateTime # we'll install this in the module context

import locale
import calendar, time
import xmlrpclib

class MetadataServiceException(Exception):
	pass

# FIXME optimize this function
def is_utf8_encoded(string):
	try: st = transcode(string,"UTF-8","UTF-8",forced = True)
	except (UnicodeDecodeError,UnicodeEncodeError): return False
	return True

def transcode_to_platform_encoding(string,source_encoding):
	""" Transcode a regular string from a source encoding to the platform's target encoding"""
	if type(string) is not str: raise ValueError
	return transcode(string,source_encoding,locale.getpreferredencoding())

def platform_to_utf8(string):
	""" This function takes a platform-encoded string and transcodes it to a plain string encoded in UTF-8 """
	return transcode(string,locale.getpreferredencoding(),"UTF-8")

def utf8_to_platform(string):
	""" This function takes an UTF-8 encoded string and transcodes it to a plain string encoded in your platform's preferred encoding """
	return transcode(string,"UTF-8",locale.getpreferredencoding())

def transcode_to_utf8_encoding(string,source_encoding):
	""" Transcode a regular string from a source encoding to the platform's target encoding"""
	if type(string) is not str: raise ValueError
	return transcode(string,source_encoding,locale.getpreferredencoding())

def transcode(string,input_encoding,output_encoding, forced = False):
	""" Transcodes a regular string from a source encoding to the target encoding """
	if type(string) is not str: raise ValueError
	if input_encoding == output_encoding and not forced: return string
	return convert_from_unicode(convert_to_unicode(string,input_encoding),output_encoding)

def convert_to_unicode(a,source_encoding=None):
	""" Converts a plain string into a Unicode object.  The string is assumed to be encoded in the source_encoding, or in the platform's encoding if None"""
	if source_encoding is None: source_encoding = locale.getpreferredencoding()
	if type(a) is str: return unicode(a,source_encoding)
	elif type(a) is unicode: return a
	else: raise ValueError

def convert_from_unicode(a,target_encoding):
	""" Converts a Unicode object into a plain string, encoded in the target_encoding """
	if type(a) is unicode: return a.encode(target_encoding)
	else: raise ValueError

def local_timestamp_to_utc(ts):
	""" Returns the equivalent UTC timestamp for the passed local timestamp"""
	return calendar.timegm(time.localtime(ts))
	
def utc_timestamp_to_local(ts):
	""" Returns the equivalent local timestamp for the passed UTC timestamp"""
	ts = int(ts)
	localtime = local_timestamp_to_utc(ts)
	offset = ts - localtime
	return ts + offset


def shell_quote(string):
	return "'%s'" % string.replace("'","'\\''")
	
import re
def remove_whitespace(text): # canonicalizes whitespace! yay!
	return re.compile("\s+").sub(" ",text).strip()

def detect_encoding(text):
	"""Detects encoding of a short piece of text.  Doesn't go beyond FIXME characters.  Can return None if no encoding could be detected."""
	# FIXME, this is a stub
	return None
	if is_utf8_encoded(text): return "UTF-8"
	else: return None
	
	
	
import types
import locale
# make the reverse function!
def to_platform_strings(params):
	"""Take an array, list, dictionary of parameters which contains strings, and convert all Unicode instances to platform-encoded strings."""
	if type(params) is list:
# 		print "stringifying list %s" % params
		return [ to_platform_strings(param) for param in params ]
	elif type(params) is tuple:
# 		print "stringifying tuple %s" % params
		return tuple([ to_platform_strings(param) for param in params ])
	elif isinstance(params,types.DictType):
# 		print "stringifying dictionary %s" % params
		resdict = {}
		for key,value in params.iteritems():
			resdict[to_platform_strings(key)] = to_platform_strings(value)
		return resdict
	elif type(params) is unicode:
# 		print "stringifying unicode %s" % params
		return convert_from_unicode(params,locale.getpreferredencoding())
	else:
# 		print "stringifying other type %s %s" % (type(params),params)
		return params


def from_platform_strings(params):
	"""Take an array, list, dictionary of parameters which contains strings, and reencodes all string instances to utf8 strings, assuming that the source encoding is the platform's."""
	if type(params) is list:
# 		print "stringifying list %s" % params
		return [ to_platform_strings(param) for param in params ]
	elif type(params) is tuple:
# 		print "stringifying tuple %s" % params
		return tuple([ to_platform_strings(param) for param in params ])
	elif isinstance(params,types.DictType):
# 		print "stringifying dictionary %s" % params
		resdict = {}
		for key,value in params.iteritems():
			resdict[to_platform_strings(key)] = to_platform_strings(value)
		return resdict
	elif type(params) is str:
# 		print "stringifying unicode %s" % params
		return convert_to_unicode(params)
	else:
# 		print "stringifying other type %s %s" % (type(params),params)
		return params


def from_xml_params(params):
	"""Take an array, list, dictionary of parameters which contains strings, and convert all Unicode instances to platform-encoded strings."""
	f = from_xml_params
	if type(params) is list: return [ f(param) for param in params ]
	elif type(params) is tuple: return tuple([ f(param) for param in params ])
	elif isinstance(params,types.DictType):
		resdict = {}
		for key,value in params.iteritems(): resdict[f(key)] = f(value)
		return resdict
	elif type(params) is unicode: return convert_from_unicode(params,locale.getpreferredencoding())
	elif isinstance(params,xmlrpclib.DateTime):
		# convert the (UTC) xmlrpclib.DateTime instance into a (UTC) mx.DateTime instance
		params = DateTime.strptime(params.value,"%Y%m%dT%H:%M:%S")
		return params
	return params


def to_xml_params(params):
	"""Take an array, list, dictionary of parameters which contains strings, and reencodes all string instances to utf8 strings, assuming that the source encoding is the platform's."""
	f = to_xml_params
	if type(params) is list: return [ f(param) for param in params ]
	elif type(params) is tuple: return tuple([ f(param) for param in params ])
	elif isinstance(params,types.DictType):
		resdict = {}
		for key,value in params.iteritems(): resdict[f(key)] = f(value)
		return resdict
	elif type(params) is str: return params
	elif type(params) is DateTimeType:
		# convert the (UTC) mx.DateTime instance into a (UTC) xmlrpclib.DateTime instance 
		params = xmlrpclib.DateTime(utc_timestamp_to_local(params.gmticks()))
		return params
	return params



import logging
from PublicInterface import PublicInterface

class Component:
	"""The base class for all metadata service components.  Implementations of Component may present a public interface (PublicInterface) to the Component constructor, via Component.__init__(public_interface=MyInterfaceImplementation).
	
	Component implementations have a private log instance, accessible at self.log, which is a Python Logger object."""

	log = None
	public_interface = None
	__component_name = None

	def __init__(self,public_interface=None):
		"""All metadata services have a name"""
		# Contract:
		assert isinstance(public_interface,PublicInterface) or public_interface is None

		# Implementation:
		self.log = logging.getLogger(self.get_component_name())
		self.public_interface = public_interface

	def get_component_name(self):
		if self.__component_name: return self.__component_name  # in the future, if a name is set, there has to be a way to change the logger's name as well...
		a = repr(self.__class__)
		a = a.split(" ")[1]
		a = a.replace("\'","").replace(">","")
		return a

	def set_component_name(self,name):
		assert type(name) is str and len(name) > 0
		self.__component_name = name
		self.log = logging.getLogger(self.get_component_name())


class Database:
	"""The interface to which all databases must conform."""

	def __init__(self):
		raise NotImplementedError

	def update_schema(self):
		raise NotImplementedError

	def open_connection(self):
		raise NotImplementedError

	def close_connection(self,connection):
		raise NotImplementedError

	def root(self,connection):
		raise NotImplementedError

	def begin_transaction(self,connection):
		raise NotImplementedError
	
	def abort_transaction(self,connection):
		raise NotImplementedError

	def commit_transaction(self,connection):
		raise NotImplementedError

	def close(self):
		raise NotImplementedError
		
	def pack(self):
		raise NotImplementedError


from Service import Service
from AdaptiveSleeper import AdaptiveSleeper
from Index import Index
from Searchable import Searchable
from Config import MetadataServiceConfig

class InfospaceAgent(Service,AdaptiveSleeper,Searchable):
	"""The base class for all metadata service infospace agents.  Agents follow a Service pattern.  Implementors need to implement the search method of the Searchable interface."""
	
	index = None
	infospace_name = None
	
	def __init__(self,index,infospace_name,config,public_interface=None):
		assert isinstance(config,MetadataServiceConfig)
		assert isinstance(index,Index)
		assert type(infospace_name) is str and len(infospace_name) > 0
		Service.__init__(self,public_interface=public_interface)
		AdaptiveSleeper.__init__(self)
		self.infospace_name = infospace_name
		self.index = index
		self.config = config



from MetadataService.InformationObject import InformationObject
from MetadataService.common import DateTime
from platform import strip_mountpoint
import os
from InfospaceObjects import Volume

class File(InformationObject):
	volume = None
	path = None
	absolute_path = None
	size = None
	modification_date = None
	mimetype = None
	
	File_catalogable = "absolute_path,size,modification_date,mimetype".split(",")

	def __init__(self,filename,mimetype,volume):
		assert isinstance(volume,Volume)
		assert type(filename) is str and len(filename) > 0
		assert type(mimetype) is str and len(mimetype) > 0
		InformationObject.__init__(self)
		self.volume = volume
		self.mimetype = mimetype
		mp = volume.get_mountpoint()
		if mp is None: raise IOError, "The file's volume vanished before its time!"
		self.path = strip_mountpoint(filename,mp)
		self.absolute_path = os.path.normpath("/%s/%s"%(volume.guid,self.path))
		statdata = os.stat(filename)
		self.size = statdata.st_size
		self.modification_date = DateTime.gmtime(statdata.st_mtime) # document this.  This construct means: construct a DateTime representation based on a UTC ticks value (same as time.time())  We should move to using Zope's DateTime instead maybe, and using DateIndexes instead of fieldindexes for that.

# commented out for the time being.  this sucks!
# 	def as_marshallable_representation(self):
# 		m = InformationObject.as_marshallable_representation(self)
# 		vol = self.volume
# 		del m["volume"]
# 		m["volume_guid"] = vol.guid
# 		m["volume_mountpoint"] = vol.get_mountpoint()
# 		fn = self.get_filename()
# 		if fn is not None: m["filename"] = fn
# 		return m
# 	
	def get_filename(self):
		mp = self.volume.get_mountpoint()
		if mp: return os.path.join(mp,self.path)

	def __str__(self):
		return "file %s (of type %s)"%(self.get_filename(),self.mimetype)

	# for the index to call before returning results, and decide what to purge itself  fixme think this through this should be in the base class
# 	def is_valid(self):
# 		if os.path.exists(self.filename): return True
# 		return False
		
	# busquemosle un uso a esto
# 	def is_fresh(self):
# 		bla bla si la vejez del archivo no corresponde a lo que esta indexado, simplemente solici



from MetadataService.InterfaceServer import InterfaceServer
from DocXMLRPCServer import DocXMLRPCServer
import ConfigParser
from MetadataService.common import from_xml_params,to_xml_params
from SimpleXMLRPCServer import SimpleXMLRPCDispatcher

class XMLRPCInterfaceServer(InterfaceServer,DocXMLRPCServer):
	""" A self-documenting implementation of the public interface server using standard TCP/HTTP transport """
	
	def __init__(self,config,interfaces):
		InterfaceServer.__init__(self,interfaces=interfaces,config=config)
		self.set_component_name("InterfaceServer.TCPXMLRPC")
		try:
			address = self.config.get("TCP interface","address")
		except (ConfigParser.NoSectionError,ConfigParser.NoOptionError):
			address = "0.0.0.0"
		try:
			port = self.config.getint("TCP interface","port")
		except (ConfigParser.NoSectionError,ConfigParser.NoOptionError):
			port = 2081

		DocXMLRPCServer.__init__(self,(address,port))
		self.set_server_title("Metadata service XML-RPC interface")
		self.set_server_name("Metadata service XML-RPC interface")
		self.set_server_title("This server maintains a list of public interfaces to the metadata service.")
		self.register_instance(self.public_interfaces)
		self.register_introspection_functions()
		self.register_multicall_functions()
		self.setDaemon(True) # fixme, this avoids the blocking upon shutdown
		
	def run(self):
		self.serve_forever()

	# FIXME: this is the place to determine user ID and submit it to the public interface
	# for that, we need to write our own request handlers
	def _dispatch(self,method,params):
		params = from_xml_params(params)
		return to_xml_params(SimpleXMLRPCDispatcher._dispatch(self,method,params))



PluginInterface.require_command("id3info")
from mx import DateTime

class MPEGAudioFile(PluginInterface.File):

	date = None
	sample_rate = None
	track_number = None
	bitrate = None
	title = None
	album = None
	genre = None
	artist = None
	# fixme implement comment and length extraction
# 	comment = None
# 	length = None

	MPEGAudioFile_catalogable = [
		"date","sample_rate","track_number","bitrate",
		"title","album","genre","artist","comment" ]

	def __init__(self,filename,mimetype,volume):
		PluginInterface.File.__init__(self,filename,mimetype,volume)
		metadata = id3info_extract(filename)
		for key,value in metadata.iteritems():
			if key == "TYER":
				self.date = DateTime.DateTime(value,1,1)
			elif key == "frequency":
				self.sample_rate = value
			elif key == "TRCK":
				self.track_number = value
			elif key == "TIT2":
				self.title = value
			elif key == "TALB":
				self.album =value
			elif key == "bitrate":
				self.bitrate = value
			elif key == "TCON":
				self.genre = value
			elif key == "COMM":
				self.comment = value
			elif key == "TPE1":
				self.artist = value

	def __str__(self):
		return "MP3 %s (%s - %s - %s)"%(self.get_filename(),self.artist,self.album,self.title)

def get_factory():
	return MPEGAudioFile

def get_mimetypes():
	return ["audio/mpeg","audio/x-mp3"]

def id3info_extract(filename):
	import commands
	output = commands.getoutput("LANG=en_US.iso-8859-1 id3info %s"%PluginInterface.shell_quote(filename))
	output = PluginInterface.transcode_to_platform_encoding(output,"iso-8859-1")
	lines = output.splitlines()
	tags = [ a[4:] for a in lines if a.startswith("===") ]
	t = {}
	for tag in tags:
		splitted = tag.split(":",1)
		key,value = splitted[0],splitted[1]
		key = key.strip().split(" ",1)[0].strip()
		value = value.strip()
		t[key] = value
	mp3info = [ a for a in lines if a.startswith("Bitrate") or a.startswith("Frequency") ]
	for a in mp3info:
		splitted = a.split(":",1)
		key,value = splitted[0],splitted[1]
		value = value.strip()
		if key == "Bitrate": t["bitrate"] = value
		if key == "Frequency": t["frequency"] = value
	
	key = "TYER"
	if t.has_key(key):
		try: t[key] = int(t[key])
		except:
			PluginInterface.log("id3lib-audio-mpeg").warning("Error decoding year (%s)",t[key],exc_info=True)
			del(t[key])
	key = "TRCK"
	if t.has_key(key):
		try: t[key] = int(t[key].split("/",1)[0])
		except:
			PluginInterface.log("id3lib-audio-mpeg").warning("Error decoding track (%s)",t[key],exc_info=True)
			del(t[key])
	key = "COMM"
	if t.has_key(key):
		if t[key].startswith("(ID3v1 Comment)["): t[key] = t[key][22:]
	key = "bitrate"
	if t.has_key(key):
		try: t[key] = int(t[key][:-4])*1000
		except: 
			PluginInterface.log("id3lib-audio-mpeg").warning("Error decoding bitrate (%s)",t[key],exc_info=True)
			del(t[key])
	key = "frequency"
	if t.has_key(key):
		try:
			val = int(t[key][:-3])
			if val == 44: t[key] = 44100
			elif val == 48: t[key] = 48000
			elif val == 24: t[key] = 24000
			elif val == 22: t[key] = 22050
			elif val == 11: t[key] = 11025
			elif val == 8: t[key] = 8000
			else: raise Exception, "Unknown bitrate value %s"%val
		except: 
			PluginInterface.log("id3lib-audio-mpeg").warning("Error decoding frequency (%s)",t[key],exc_info=True)
			del(t[key])
	
	def get_genre_from_int(genrenr):  # inner function with the tables
		assert type(genrenr) is int
		if not hasattr(get_genre_from_int,"genres"): get_genre_from_int.genres = id3_builtin_genres
		if get_genre_from_int.genres.has_key(genrenr): return get_genre_from_int.genres[genrenr]
	
	key = "TCON"
	if t.has_key(key):
		try:
			if t[key].startswith("(") and t[key].endswith(")"):
				genrenr = int (t[key][1:-1])
				genrename = get_genre_from_int(genrenr)
				if genrename is not None: t[key] = genrename
				else:
					PluginInterface.log("id3lib-audio-mpeg").warning("Error decoding genre (%s)",t[key],exc_info=True)
					del(t[key])
		except:
			PluginInterface.log("id3lib-audio-mpeg").warning("Error decoding genre (%s)",t[key],exc_info=True)
			del(t[key])
			
	return t

id3_builtin_genres = {
0 : "Blues", 
1 : "Classic Rock", 
2 : "Country", 
3 : "Dance", 
4 : "Disco", 
5 : "Funk", 
6 : "Grunge", 
7 : "Hip-Hop", 
8 : "Jazz", 
9 : "Metal", 
10 : "New Age", 
11 : "Oldies", 
12 : "Other", 
13 : "Pop", 
14 : "R&B", 
15 : "Rap", 
16 : "Reggae", 
17 : "Rock", 
18 : "Techno", 
19 : "Industrial", 
20 : "Alternative", 
21 : "Ska", 
22 : "Death Metal", 
23 : "Pranks", 
24 : "Soundtrack", 
25 : "Euro-Techno", 
26 : "Ambient", 
27 : "Trip-Hop", 
28 : "Vocal", 
29 : "Jazz+Funk", 
30 : "Fusion", 
31 : "Trance", 
32 : "Classical", 
33 : "Instrumental", 
34 : "Acid", 
35 : "House", 
36 : "Game", 
37 : "Sound Clip", 
38 : "Gospel", 
39 : "Noise", 
40 : "AlternRock", 
41 : "Bass", 
42 : "Soul", 
43 : "Punk", 
44 : "Space", 
45 : "Meditative", 
46 : "Instrumental Pop", 
47 : "Instrumental Rock", 
48 : "Ethnic", 
49 : "Gothic", 
50 : "Darkwave", 
51 : "Techno-Industrial", 
52 : "Electronic", 
53 : "Pop-Folk", 
54 : "Eurodance", 
55 : "Dream", 
56 : "Southern Rock", 
57 : "Comedy", 
58 : "Cult", 
59 : "Gangsta", 
60 : "Top 40", 
61 : "Christian Rap", 
62 : "Pop/Funk", 
63 : "Jungle", 
64 : "Native American", 
65 : "Cabaret", 
66 : "New Wave", 
67 : "Psychadelic", 
68 : "Rave", 
69 : "Showtunes", 
70 : "Trailer", 
71 : "Lo-Fi", 
72 : "Tribal", 
73 : "Acid Punk", 
74 : "Acid Jazz", 
75 : "Polka", 
76 : "Retro", 
77 : "Musical", 
78 : "Rock & Roll", 
79 : "Hard Rock", 
80 : "Folk", 
81 : "Folk-Rock", 
82 : "National Folk", 
83 : "Swing", 
84 : "Fast Fusion", 
85 : "Bebob", 
86 : "Latin", 
87 : "Revival", 
88 : "Celtic", 
89 : "Bluegrass", 
90 : "Avantgarde", 
91 : "Gothic Rock", 
92 : "Progressive Rock", 
93 : "Psychedelic Rock", 
94 : "Symphonic Rock", 
95 : "Slow Rock", 
96 : "Big Band", 
97 : "Chorus", 
98 : "Easy Listening", 
99 : "Acoustic", 
100 : "Humour", 
101 : "Speech", 
102 : "Chanson", 
103 : "Opera", 
104 : "Chamber Music", 
105 : "Sonata", 
106 : "Symphony", 
107 : "Booty Bass", 
108 : "Primus", 
109 : "Porn Groove", 
110 : "Satire", 
111 : "Slow Jam", 
112 : "Club", 
113 : "Tango", 
114 : "Samba", 
115 : "Folklore", 
116 : "Ballad", 
117 : "Power Ballad", 
118 : "Rhythmic Soul", 
119 : "Freestyle", 
120 : "Duet", 
121 : "Punk Rock", 
122 : "Drum Solo", 
123 : "A capella", 
124 : "Euro-House", 
125 : "Dance Hall", 
126 : "Goa", 
127 : "Drum & Bass", 
128 : "Club-House", 
129 : "Hardcore", 
130 : "Terror", 
131 : "Indie", 
132 : "Britpop", 
133 : "Negerpunk", 
134 : "Polsk Punk", 
135 : "Beat", 
136 : "Christian Gangsta Rap", 
137 : "Heavy Metal", 
138 : "Black Metal", 
139 : "Crossover", 
140 : "Contemporary Christian", 
141 : "Christian Rock ", 
142 : "Merengue", 
143 : "Salsa", 
144 : "Trash Metal", 
145 : "Anime", 
146 : "JPop", 
147 : "Synthpop",
}



#!/usr/bin/env python

# FIXME standardize on a common response/faultCode format and common codes

class PublicInterface:
	"""The public interface for your component. # fixme clarify this docs
	
	To publish a set of functions for your component, you implement the PublicInterface class and in your component's constructor, call Component.__init__(MyInterface(component=self)).  In any of your implementation's methods, you will have the self.__component attribute available for you to operate on your component.

	In your implementation of the interface, each public (non _) method and object is available for the whole world to use through the interface servers.  When the user requests a method through an interface server, the interface server automatically locates the appropriate method to run, passing the user ID (which can be None) and the user-supplied parameters to your method."""

	_component = None
	_name = None
	
	def __init__(self,name,component):
		"""The default implementation automatically assigns the passed component to self._component"""
		#contract
		from Component import Component
		assert isinstance(component,Component)
		assert type(name) is str and len(name) > 0
		#implementation
		self._component = component
		self._name = name
		
	def get_name(self):
		return self._name

	def get_component(self): # proteger esta funcion de outside accessors
		return self._component

#FIXME validate each request per client UID like if uid != 0 (it could be none coming from the web interface, e.g.) then raise something which notifies the client that you're fucked because you're not root
# class AdministrativeInterface:
# 	""" An abstraction encapsulating admin-only requests to the metadata service """
# 	__metadata_service = None
#
# 	def __init__(self,metadata_service):
# 		self.__metadata_service = metadata_service
# 		self.__log = logging.getLogger("interfaces.administrative")
#
# 	def exit(self):
# 		""" Terminate the metadata service cleanly """
# 		self.__log.debug("Exit request on public interface")
# 		self.__metadata_service.request_exit()
# 		return 0
#
# 	def reloadConfig(self):
# 		""" Reload configuration file and plugins  """
# 		self.__log.debug("Reload configuration request on public interface")
# 		self.__metadata_service.request_reload_config()
# 		return 0
#
# # 	def getCacheSize(self):
# # 		""" Get the cache size"""
# # 		self.__log.debug("Cache size request on public interface")
# # 		return self.__metadata_service.index_database.getCacheSize()
# #
# # 	def getCacheDetailSize(self):
# # 		""" Get the cache size, detailed by connection """
# # 		self.__log.debug("Cache detail size request on public interface")
# # 		return self.__metadata_service.index_database.cacheDetailSize()
# #
# # 	def getObjectCount(self):
# # 		""" Get the number of stored objects """
# # 		self.__log.debug("Object count request on public interface")
# # 		return self.__metadata_service.index_database.objectCount()
# #
# # 	def minimizeCache(self,size):
# # 		""" Get the number of stored objects """
# # 		self.__log.debug("Cache minimize request on public interface")
# # 		self.__metadata_service.index_database.cacheMinimize(size)
# # 		return 0
#
# #FIXME not here! create a subclass of DatabaseOperation named DatabaseModificationOperation which, when passed to the execute method, shall create a new transaction and run the operation in that context.  Semantics for databaseoperation should also be changed.  Document the semantics and document the fact that if the user wants a transaction to be aborted, all he has to do is simply raise an exception (check that's the case!)
# #FIXME not here, but rather in the HTML plugin , the KDE metainfo is returning fucked up strings for html document titles, better to pull the title out with a regular expression!
# #FIXME to return faults we can use a Fault instance defined somewhere in SimpleXMLRPCServer!
# #FIXME document the fact that all methods need to return at least one non-None value, such as "OK" or 0
#
# class MetadataServicePublicInterface(PublishableInterface):
# 	""" This is the official public interface, which aggregates all other interfaces. """
# 	index = None
# 	administration = None
#
# 	def __init__(self,metadata_service):
# 		self.index = IndexInterface(metadata_service)
# 		self.administration = AdministrativeInterface(metadata_service)
#



# import commands
import logging
import re

class HTMLDocument(PluginInterface.File):

	title = None
	contents = None
	encoding = None

	HTMLDocument_catalogable = [ "title","contents","encoding" ]

	def __init__(self,filename,mimetype,volume):
		PluginInterface.File.__init__(self,filename,mimetype,volume)
		
		f = file(filename,"rb")
		output = f.read(-1)
		f.close()
			
		filerepr = HTMLFile(output)

		try:
			encoding = filerepr.get_encoding()
			utftitle = filerepr.get_title_utf8()
			utfbody = filerepr.get_plaintext_utf8()
		except (TypeError,LookupError):
			PluginInterface.log("text-html").info("%s's declared encoding %s is invalid",filename,repr(encoding))
			return
			

		try:
			self.encoding = encoding
			if utftitle: self.title = PluginInterface.utf8_to_platform(utftitle)
			if utfbody: self.contents = PluginInterface.utf8_to_platform(utfbody)
		except (UnicodeDecodeError, UnicodeEncodeError):
			PluginInterface.log("text-html").info("%s is in an encoding %s that is unsupported by the platform",filename,repr(encoding),exc_info=True)


	def __str__(self):
		return "HTML document %s (%s)"%(self.get_filename(),self.title)


import htmlentitydefs
class HTMLFile:

	title = None
	sent_markup = None
	utf8_markup = None
	encoding = None
	utf8_body = None
	plaintext = None
	
	def __init__(self,markup):
		if type(markup) is not str:
			raise TypeError, "HTMLFiles can only be instanced with regular strings"
		self.sent_markup = markup
		# save the markup
		
	def __process(self):
		# detect encoding  DONE
# 		encoding = self.get_encoding()
		# transcode everything to UTF-8
# 		utf8_markup = self.get_utf8_markup()
		# get the body
		body,title = self.get_body_and_title()
		# remove all tags, putting spaces wherever was one
		# replace all entities for their Unicode codepoints
		def strip_scripts(text):
			return re.compile("<script.*?>.*?</script.*?>",re.DOTALL|re.IGNORECASE).sub("",text)
		def strip_comments(text):
			return re.compile("<!.*?-->",re.DOTALL|re.IGNORECASE).sub("",text)
		def strip_tags(text):
			return re.compile("<.*?>",re.DOTALL).sub("",text)
		def replace_entities(text):
			def convert_codepoint_to_utf8_char(codepoint):
				cp = str(codepoint)
				return ("\u" + "0" * (4 - len(cp)) + cp).decode("unicode_escape").encode("UTF-8")
			def map_entity_to_codepoint(entity):
				import htmlentitydefs
				entity = entity.group(1)
				if entity[0] == "#": return convert_codepoint_to_utf8_char(entity[1:])
				else:
					d = htmlentitydefs.name2codepoint
					if d.has_key(entity[1:]): return convert_codepoint_to_utf8_char(d[entity[1:]])
				return ""
			return re.compile("&(.*?);").sub(map_entity_to_codepoint,text)
		if body:
			bodytext = body
			bodytext = strip_comments(bodytext)
			bodytext = strip_scripts(bodytext)
			bodytext = strip_tags(bodytext)
			bodytext = replace_entities(bodytext)
			bodytext = PluginInterface.remove_whitespace(bodytext)
			body = bodytext
			self.plaintext = body
		if title:
			titletext = title
			titletext = strip_tags(titletext)
			titletext = replace_entities(titletext)
			titletext = PluginInterface.remove_whitespace(titletext)
			title = titletext
			self.title = title
		# save the final text in a member variable

	def get_utf8_markup(self):
		if self.utf8_markup is None:
			original_encoding = self.get_encoding()
			self.utf8_markup = PluginInterface.transcode(self.sent_markup,original_encoding,"UTF-8")
		return self.utf8_markup
	
	def get_declared_encoding(self):
		# return the file's declared encoding
		def get_charset_regexp():
			charset="charset"
			igual="="
			definit = ".*?"
			comilla = "[" + "'" + '"' + "]"
			expresion = charset + igual + "(%s)"%definit + comilla
			return expresion
		def get_encoding_regexp():
			encoding="encoding"
			igual="="
			definit = ".*?"
			comilla = "[" + "'" + '"' + "]"
			expresion = encoding + igual + comilla + "(%s)"%definit + comilla
			return expresion
		
		objeto = re.compile(get_encoding_regexp(),re.IGNORECASE)
		cuadre = objeto.search(self.sent_markup)
		if cuadre:
			return cuadre.group(1).strip()
		objeto = re.compile(get_charset_regexp(),re.IGNORECASE)
		cuadre = objeto.search(self.sent_markup)
		if cuadre:
			return cuadre.group(1).strip()

	def get_encoding(self):
		if self.encoding is None:
		
			# auto-detect encoding
			self.encoding = self.get_declared_encoding()
			if self.encoding is None:
				encoding = PluginInterface.detect_encoding(self.sent_markup)
				if encoding: self.encoding = encoding
				else: self.encoding = "ISO-8859-1" #fallback!
		
		return self.encoding
		
	def get_body_and_title(self):
		utf8_markup = self.get_utf8_markup()
		bodyobject = re.compile("<body.*?>(.*)</body.*?>",re.DOTALL | re.IGNORECASE )
		bodymatch = bodyobject.search(utf8_markup)
		titleobject = re.compile("<title.*?>(.*)</title.*?>",re.DOTALL | re.IGNORECASE )
		titlematch = titleobject.search(utf8_markup)
		if bodymatch: body = bodymatch.group(1).strip()
		else: body = ""
		if titlematch: title = titlematch.group(1).strip()
		else: title = ""
		
		return (body,title)
		
	def get_title_utf8(self):
		if self.title is None:
			self.__process()
		return self.title
		
	def get_plaintext_utf8(self):
		if self.plaintext is None:
			self.__process()
		return self.plaintext



def get_factory():
	return HTMLDocument

def get_mimetypes():
	return ["text/html"]



import time
from Service import Service

class AdaptiveLoadController(Service):
	"""The AdaptiveLoadController is a separate thread which periodically checks the system load and adapts the sleep time for AdaptiveSleeper-inherited threads.  Do not instantiate directly: use the get_adaptive_load_controller singleton factory."""
	throttle_ms = 0.0
	error_margin = None
	target_load = None
	max_throttling = None
	sleeper_granularity = 0.5 # this is because you might actually start a sleeper without having started the controller yet
	check_frequency = None

	def __init__(self,target_load=1,error_margin=0.1,max_throttling=5,sleeper_granularity=0.5,check_frequency=1.0):
		""" Initializes the load controller.  target_load is the maximum load before starting to throttle, error_margin is the allowed +- margin on the load, max_throtttling is the maximum number of seconds Sleeper threads should sleep, sleeper_granularity is the granularity with which the Sleepers check if it's time to continue, and check_frequency is the time between recalculations of the throttling value"""
		Service.__init__(self,public_interface=AdaptiveLoadControllerInterface("loadController",self))
		self.set_component_name("AdaptiveLoadController")
		self.error_margin = error_margin
		self.target_load = target_load
		self.max_throttling = max_throttling
		AdaptiveLoadController.sleeper_granularity = sleeper_granularity
		self.check_frequency = check_frequency

	def set_target_load(self,target_load):
		assert type(target_load) is float
		if target_load <= 0.09 or target_load > 10:
			raise ValueError, "target_load must be between 0.1 and 10"
		self.target_load = target_load
	
	def run(self):
		while not self.killed():
			self.recalculate_throttling()
			time.sleep(self.check_frequency)
		AdaptiveLoadController.throttle_ms = 0.0

	def recalculate_throttling(self):

		old_throttle_ms = AdaptiveLoadController.throttle_ms
		def inrange(number,target,plusminus):
			return (number < target +plusminus and number > target - plusminus)

		import os
		load= os.getloadavg()[0]
		if inrange(load,self.target_load,self.error_margin):
			pass
		elif load > self.target_load:
			if AdaptiveLoadController.throttle_ms < self.max_throttling:
				n = AdaptiveLoadController.throttle_ms + (load - self.target_load)*10
				AdaptiveLoadController.throttle_ms = min(self.max_throttling,n)
		elif load < self.target_load:
			if AdaptiveLoadController.throttle_ms > 0.0:
				n = AdaptiveLoadController.throttle_ms - (self.target_load - load)*10
				AdaptiveLoadController.throttle_ms = max(0.0,n)
				
		if old_throttle_ms != AdaptiveLoadController.throttle_ms:
			self.log.debug("Target load %s, current load %s, throttle at %s seconds",self.target_load,load,AdaptiveLoadController.throttle_ms)
		

from PublicInterface import PublicInterface

class AdaptiveLoadControllerInterface(PublicInterface):

	def setTargetLoad(self,target_load):
		target_load = float(target_load)
		self.get_component().set_target_load(target_load)
		return True


import ZODB
from persistent.mapping import PersistentMapping
from persistent import Persistent
from BTrees.IOBTree import IOBTree
from BTrees.OOBTree import OOBTree
from BTrees.OIBTree import OIBTree
from MetadataService.InformationObject import InformationObject
from VolumeExpert import VolumeExpert
from MetadataService.Infospace import Infospace
#from Products.ZCatalog.Catalog import Catalog
#from Products import PluginIndexes
#from Products.TextIndexNG2 import TextIndexNG
from MetadataService.common import DateTimeType
import time
import logging

class VolumeList(PersistentMapping):

	def has_volume(self,guid):
		assert type(guid) is str and len(guid) > 0
		return self.has_key(guid)
	
	def get_volume(self,guid):
		return self[guid]
	
	def has_volume(self,guid):
		return self.has_key(guid)
	
	def store_volume(self,volume):
		if self.has_volume(volume.guid): self.remove_volume(volume.guid)
# 		volume._volume_list = self
		self[volume.guid] = volume
	
	def remove_volume(self,guid):
		if self.has_volume(guid):
			volume = self.get_volume(guid)
			volume.remove_all_files()
			del self[guid]
			
	def __init__(self):
		PersistentMapping.__init__(self)
		# initialize catalogs
# 		self.date_catalog = DateCatalog()
#		self.textindexng2_catalog = TextIndexNG2Catalog()
#		self.field_catalog = FieldCatalog()
#		self.path_catalog = PathCatalog()

	# this is called by the volume
	def _catalog_file(self,volume,file):
		cat_key = "|".join([volume.guid,file.path])
		all_involved_catalogs = {}
		for attrname,value in file.__dict__.iteritems():
			c = []
			if attrname[0:1] == "_": continue
			if type(value) is DateTimeType: c = [self.field_catalog]
# 			elif attrname is "volume": c = [self.field_catalog] # uncomment when fieldindex bug is fixed
			elif attrname is "contents": c = [self.textindexng2_catalog]
			elif attrname is "absolute_path": c = [self.path_catalog,self.field_catalog]
			elif attrname is "path": c = []
			elif type(value) is str: c = [self.field_catalog,self.textindexng2_catalog] #,self.textindexng2_catalog
			elif type(value) in (int,long): c = [self.field_catalog]
			for involved_catalog in c:
				all_involved_catalogs[str(involved_catalog)] = involved_catalog
				involved_catalog.make_index(attrname)
		for i in all_involved_catalogs.itervalues():
			t = time.time()
			i.catalogObject(object=file,uid=cat_key)
			logging.getLogger("Volume."+volume.guid).debug("Catalogued %s in %s in %s sec",file,i,time.time()-t)
			
	
	def _uncatalog_file(self,volume,cat_key):
		cat_key = "|".join([volume.guid,cat_key])
		catalogs = [self.field_catalog,self.path_catalog,self.textindexng2_catalog]#self.date_catalog,
		for i in catalogs:
			t = time.time()
			i.uncatalogObject(uid=cat_key)
			logging.getLogger("Volume."+volume.guid).debug("Uncatalogued %s from %s in %s",cat_key,i,time.time() -t)

	def search(self,expression): #FIX
		assert isinstance(expression,Expression)
		
		if isinstance(expression,Aggregation):
			resultset = None
			for e in expression.expressions:
				er = self.search(e)
				 # why bother, intersection with empty will always yield empty
				if isinstance(expression,And) and not er: return er
				logging.getLogger("Search").debug("%s results for expression %s",len(er),e)
				if resultset is None: resultset = er
				else:
					if isinstance(expression,Or): resultset.union_update(er)
					elif isinstance(expression,And):
						resultset.intersection_update(er)
						 # the intersection yielded an empty set, abort further operations
						if not resultset: return resultset
			if resultset is None: raise ValueError, "Aggregation expressions must contain at least one criterion expression"
			return resultset
			
		elif isinstance(expression,Criterion):
			
			# choose the right catalog based on query type
			opclass = expression.operator.__class__
			if opclass in [Equals,IsEqualOrLessThan,IsEqualOrGreaterThan]:
				right_catalog = self.field_catalog
			elif opclass in [IsBelow,IsOrIsBelow]:
				right_catalog = self.path_catalog
			elif opclass in [NL]:
				right_catalog = self.textindexng2_catalog
			else: raise NotImplementedError, "%s queries on %s not available"%(str(expression.operator),expression.attribute)
			
			#check that the chosen catalog to query has an index for that attribute and can, in fact, be queried for that attribute
			index_name = expression.attribute
			if not right_catalog.has_index(index_name):
				raise NotImplementedError, "no index on %s for attribute %s"%(right_catalog,index_name)
			
			# construct the proper query based on query type and attribute...
			if opclass is Equals: searchexp = {expression.attribute: { "query" : expression.value } }
			elif opclass is NL:
				searchexp = {expression.attribute: { "query" : expression.value} }
			elif opclass is IsEqualOrGreaterThan:
				searchexp = {expression.attribute: { "query" : expression.value,"range":"min"} }
			elif opclass is IsEqualOrLessThan:
				searchexp = {expression.attribute: { "query" : expression.value,"range":"max"} }
			elif opclass is IsOrIsBelow:
				searchexp = {expression.attribute: { "query" : expression.value } }
			else: raise NotImplementedError, "query type %s not implemented for %s"%(expression.operator,expression.attribute)
			
			
			# ...and query the catalog do the query and make a set out of the results
			logging.getLogger("Search").debug("Searching %s for catalog expression %s",right_catalog,searchexp)

			results = right_catalog.search(searchexp)
			resultset = [ right_catalog.paths[result.getRID()] for result in results ]
			return Set(resultset) # return a set of keys
		
		else: raise Exception, "What the hell?"

#class VolumeCatalog(Catalog):
#	
#	def __init__(self):
#		Catalog.__init__(self)
#		self.aq_parent = self
#	
#	def has_index(self,attrname):
#		return self.indexes.has_key(attrname)
#		
#	def make_index(self,attrname):
#		assert type(attrname) is str and len(attrname) > 0
#		if self.has_index(attrname): return
#		idx = self._get_index(attrname)
#		logging.getLogger("VolumeCatalog").debug(
#				"Making %s index for %s",str(idx),attrname)
#		self.addIndex(attrname,idx)
#	
#	def _get_index(self,attrname):
#		raise NotImplementedError

#class FieldCatalog(VolumeCatalog):
#	def _get_index(self,attrname): return PluginIndexes.FieldIndex.FieldIndex.FieldIndex(attrname)
#class PathCatalog(VolumeCatalog):
#	def _get_index(self,attrname): return PluginIndexes.PathIndex.PathIndex.PathIndex(attrname)
#import locale
#class TextIndexNG2Catalog(VolumeCatalog):
#	def _get_index(self,attrname):
#		options = { "default_encoding":locale.getpreferredencoding()}
#		return TextIndexNG.TextIndexNG(attrname,extra=options)
# class DateCatalog(VolumeCatalog):
# 	def _get_index(self,attrname): return PluginIndexes.DateIndex.DateIndex.DateIndex(attrname)


class FilesystemInfospace(Infospace):
	volumes = None
	
	def __init__(self):
		self.volumes = VolumeList()
	
	def __str__(self):
		return "filesystem infospace, containing volumes %s" % ", ".join([ str(a) for a in self.volumes.values() ])
		
		

class FileList(Persistent):
	
# 	volume = None
# 	catalog = None
# 	highest_file_id = None
	mapping = None
	
	def keys(self): return [a for a in self.mapping.keys()]
	
	def __init__(self):
# 		assert isinstance(volume,Volume)
# 		self.volume = volume
		self.mapping = OOBTree()
		self.mod_dates = OIBTree()
		
	def store_file(self,file):
		from File import File
		assert isinstance(file,File)
		self.__save_file(file)
	
	def __save_file(self,file): # private accesor for the catalog stuff
		self.mapping[file.path] = file
		self.mod_dates[file.path] = int(file.modification_date)
	
	def get_file(self,path):
		return self.mapping[path]

	def get_file_mod_date(self,path):
	# returns the file's mod date as an int.  optimized for performance.
		return self.mod_dates[path]

	def has_file(self,path):
		return self.mapping.has_key(path)
		
	def remove_file(self,path):
		del (self.mapping[path])
		del (self.mod_dates[path])

from sets import Set
#import ExtensionClass

#class Volume(InformationObject,ExtensionClass.Base):
class Volume(InformationObject):
	guid = None
	label = None
	filesystem_type = None
	removable = None
	fully_indexed = False # accesors for this? make it private?
	file_list = None
	mountpoint = None

	def __init__(self,guid,mountpoint,filesystem_type,label=None,removable=False):
		InformationObject.__init__(self)
		assert type(guid) is str and len(guid) > 0
		assert type(filesystem_type) is str and len(filesystem_type) > 0
		assert type(removable) is bool
		self.file_list = FileList()
		self.guid = guid
		self.label = label
		self.removable = removable
		self.filesystem_type = filesystem_type
		self.fully_indexed = False
		self.mountpoint = mountpoint
		
	
	def store_file(self,file):
		if self.has_file(file.path): self.remove_file(file.path)
		self.file_list.store_file(file)
		if hasattr(self,"_volume_list"): self._volume_list._catalog_file(self,file)
	
	def remove_file(self,path):
		if self.has_file(path):
			if hasattr(self,"_volume_list"): self._volume_list._uncatalog_file(self,path)
			self.file_list.remove_file(path)
		else: raise KeyError, "no such file %s in volume %s"%(path,self)
		
	def remove_all_files(self):
		all_files_paths =self.file_list.keys()
		for path in all_files_paths:
			logging.getLogger("Volume."+self.guid).debug("Removing %s",path)
			self.remove_file(path)
		
	def has_file(self,path): return self.file_list.has_file(path)
	
	def has_file_changed_since(self,path,time_arg):
		# receives a DateTime and compares it with the file's cached mod date
		stored_mod_date = self.file_list.get_file_mod_date(path)
		if type(time_arg) is not int: dtarg = int(time_arg)
# 		logging.getLogger("has_file_changed_since").debug("time_arg %s stored_mod_date %s, changed %s",time_arg,stored_mod_date,time_arg == stored_mod_date)
		# this is not strictly correct but we expect this to be better performant
		return time_arg == stored_mod_date
	
	def get_file(self,path): return self.file_list.get_file(path)
		
	def get_mountpoint(self):
		return self.mountpoint
	
	def update_mountpoint(self,mountpoint):
		self.mountpoint = mountpoint

	def is_mounted(self):
		return self.get_mountpoint() is not None
	
# 	def __hash__(self):
# 		akku = 0
# 		for char in self.guid:
# 			akku = akku + ord(char)
# 		return akku
		
	def __str__(self):
		mountpoint = self.get_mountpoint()
		if mountpoint: mountpoint = "mounted at %s"%mountpoint
		else: mountpoint = "unmounted"
		if self.label: label = " (labeled %s)"%self.label
		else: label = ""
		return "volume %s%s, %s"%(self.guid,label,mountpoint)
	
	def __repr__(self): return self.__str__()



import platform
from MetadataService.Component import Component
import os
from platform import is_or_is_below

def is_valid_filename_encoding(filename):
	# Check whether the filename is valid-encoded
	import locale
	from MetadataService.common import transcode
	enc = locale.getpreferredencoding()
	try: transcode(filename,enc,enc,forced=True)
	except: return False
	return True


class _VolumeExpert(Component):

	user_stoplist = None

	def __init__(self,stoplist):
		assert type(stoplist) is list
		Component.__init__(self,public_interface=None)
		self.set_component_name("VolumeExpert")
		self.user_stoplist = stoplist
		self.rebuild_caches()
		
		
# 	def get_mounted_mountpoints(self):
		""" Returns a list of mountpoints for each mounted volume."""
# 		return get_mounted_mountpoints()
	
	def is_appropriate(self,filename):
		""" Returns True if a filename should be indexed.  Use this in the crawlers and wherever you need to confirm whether to pass a file to the FilesystemCrawler queues or not. """
		
		# files/dirs/volume mountpoints in stoplist or database path not appropriate
		for path in self.cache_stoplist:
			if is_or_is_below(filename,path): return False
		# backup files and directories not appropriate
		if filename.endswith('~') or '~/' in filename: return False
		# hidden directories not appropriate
		if '/.' in filename: return False
		# symlinks not appropriate
		if os.path.islink(filename): return False
		# invalid encoding is not appropriate as well
		if not is_valid_filename_encoding(filename):
			self.log.warning("%s has an invalid encoding", filename)
			return False
		
		if os.path.isfile(filename) or os.path.isdir(filename): return True # this should exclude all sockets, fifos and device nodes
		return False
		
# 	def get_appropriate_mountpoints(self):
# 		return [ l for l in self.get_mounted_mountpoints() if self.is_appropriate(l) ]

	cache_guid_to_mountpoint = {}
	cache_currently_known_volumes = {}
	cache_stoplist = []
	
	def rebuild_caches(self):
		builtin_stoplist = ["/tmp","/var/tmp","/usr/tmp","/dev"]
 		self.cache_stoplist = builtin_stoplist + self.user_stoplist
		self.cache_currently_known_volumes = platform.get_currently_known_volumes()
		self.cache_guid_to_mountpoint = {}
		mtab = platform.get_mtab()
		for guid,vr in self.cache_currently_known_volumes.iteritems():
			for record in mtab.records:
				if record.device_file == vr.current_device_file:
					self.cache_guid_to_mountpoint[guid] = record.mountpoint

	def get_mountpoint(self,guid):
		assert type(guid) is str and len(guid) > 0
		### return the mountpoint for a volume GUID
		try: return self.cache_guid_to_mountpoint[guid]
		except KeyError: return None
		
	def get_guid_for_mountpoint(self,mountpoint):
		assert type(mountpoint) is str and len(mountpoint) > 0
		### return the mountpoint for a volume GUID
		for guid,mp in self.cache_guid_to_mountpoint.iteritems():
			if mp == mountpoint: return guid
		return None
		
	def get_currently_known_volumes(self):
		return self.cache_currently_known_volumes
		
		
def VolumeExpert(stoplist):
	if not VolumeExpert.__dict__.has_key("instance"):
		VolumeExpert.instance = _VolumeExpert(stoplist)
	return VolumeExpert.instance



class Transaction: pass

from InformationObject import InformationObject

class IndexContext:

	__database = None
	__connection = None
	__catalog = None
	__catalog_connection = None
	
	def __init__(self,index):
		# Contract
		assert isinstance(index,Index)
		self.__index = index
		self.__database = index.get_database()
		self.__connection = self.__database.open_connection()
		self.__catalog = index.get_catalog()
	
	def __del__(self):
# 		self.__database.abort_transaction(self.__connection)
		if self.__connection: self.__database.close_connection(self.__connection)
		if self.__catalog_connection: self.__catalog_connection.close()
		self.__connection = None
		self.__database = None
		self.__index = None
		self.__catalog_connection = None
	
	def release(self):
		# while this may be automatically garbage-collected, it's better to release it manually once it's no longer needed
		self.abort_transaction() # releasing a context implicitly discards all changes in the objects connected to the context
		if self.__catalog_connection: self.__catalog_connection.abort_transaction()
		self.__del__()
		
	def root(self):
		return self.__database.root(self.__connection)
	
	def set_catalog_infospace(self,name):
		# FIXME should actually call the catalog daemon instead of specifying the db name here
		if self.__catalog_connection: self.__catalog_connection.close()
		self.__catalog_connection = self.__catalog.get_catalog_infospace(name).conn()

	def catalog_object(self,object_key,object):
		if self.__catalog_connection: self.__catalog_connection.catalogObject(object_key,object)
		else: raise LookupError,"Cannot catalog an object without knowing the catalog infospace first"

	def uncatalog_object(self,object_key):
		if self.__catalog_connection: self.__catalog_connection.uncatalogObject(object_key)
		else: raise LookupError,"Cannot uncatalog an object without knowing the catalog infospace first"

	def search(self,expression):
		if self.__catalog_connection: return self.__catalog_connection.search(expression)
		else: raise LookupError,"Cannot search without knowing the catalog infospace first"

	def begin_transaction(self):
		"""Requests that a transaction be started on the connection's context."""
		self.__database.begin_transaction(self.__connection)
		if self.__catalog_connection: self.__catalog_connection.begin_transaction()

# 		self.transaction = Transaction()

	def abort_transaction(self):
		"""Aborts the transaction in the context"""
		self.__database.abort_transaction(self.__connection)
		if self.__catalog_connection: self.__catalog_connection.abort_transaction()
# 		del(self.transaction)

	def commit_transaction(self,message=None):
		"""Commits the transaction in the context"""
		self.__database.commit_transaction(self.__connection)
		if self.__catalog_connection: self.__catalog_connection.commit_transaction()
# 		del(self.transaction)



from common import MetadataServiceException
class IndexException(MetadataServiceException): pass
class TransactionRequiredException(IndexException): pass


from Database import Database
from Component import Component
from pgtools.Catalog import PGCatalog

class Index(Component):

	__database = None
	__catalog = None
	infospace_agents = {}
	
	def __init__(self,database,catalog):
		# contract
		assert isinstance(database,Database)
		assert isinstance(catalog,PGCatalog)
		
		# implementation
		Component.__init__(self,public_interface=IndexInterface(component=self,name="index"))
		self.set_component_name("Index")
		self.__database = database
		self.__database.update_schema()
		self.__catalog = catalog
		self.__catalog.update_schema()

	def register_infospace_agent(self,infospace_agent):
		from InfospaceAgent import InfospaceAgent
		assert isinstance(infospace_agent,InfospaceAgent)
		self.infospace_agents[infospace_agent.infospace_name] = infospace_agent
	
	def acquire_context(self):
		"""Acquire a context to operate with the database. Contexts are automatically released when their reference count drops to zero, so don't forget to unreference them, or explicitly release them with their release() method."""
		return IndexContext(index=self)
	
	def get_database(self): return self.__database
	
	def get_catalog(self): return self.__catalog

	def close(self):
		self.__database.close()
		self.__catalog.close()
		
	def pack(self):
		self.__database.pack()
		self.__catalog.pack()


from PublicInterface import PublicInterface
from pgtools.query import normalize_query
from common import transcode
import locale
import logging
import time

class IndexInterface(PublicInterface):
	""" An abstraction encapsulating possible user-level requests to the search engine."""
	
	def pack(self):
		t = time.time()
		index = self.get_component()
		logging.getLogger("IndexInterface").debug("packing database")
		index.pack()
		logging.getLogger("IndexInterface").debug("packed database in %s sec",time.time()-t)
		return True
		

	def search(self,query):
		logging.getLogger("IndexInterface").debug("searching with query %s",query)
		""" Delegate search to infospace agents, collate results and return them """
		# this should: 1) determine which infospaces are alluded, 2) compose a Query object for each of the alluded infospaces, 3) submit the query to each infospace, 4) aggregate each infospace's result in a single result set, 5) and submit the result set back to the client
		import time
		thetime = time.time()
		context = None
		try:
			query = normalize_query(query) # normalize_Query or some other function NEEDS to transform UTF-8 encoded data and dates to the internal representations (platform-encoded strings and whatever may be the flavor of the dates' timezones I choose this month).
			
			index = self.get_component()
			context = index.acquire_context()
			
			# compute the involved infospace list
			# and prune the infospace list from those that don't support searches
			if len(query.infospaces) == 0: infospaces = index.infospace_agents.values()
			else:
				infospaces = []
				for iname in query.infospaces:
					try:
						if hasattr(index.infospace_agents[iname],"search"):
							infospaces.append(index.infospace_agents[iname])
					except KeyError: raise LookupError, "Unknown infospace %s"%iname
			
			objects = []
			
			for i in infospaces:
				results = None
				# infospace must return a list of URLs, pre-encoded in UTF-8
				results = i.search(expression = query.expression, context = context)
				if results is not None:
					assert type(results) is list
					objects.extend(results)
			
			results = objects
			
			logging.getLogger("IndexInterface").debug("Search returned %s results after %s sec. (raw query time)", len(results), time.time() - thetime)
			thetime = time.time()
# 			results = [ object.as_marshallable_representation() for object in objects ]
# 			logging.getLogger("IndexInterface").debug("Search marshalled objects after %s sec.", time.time() - thetime)
			context.release()
		
		except:
			if context: context.release()
			logging.getLogger("IndexInterface").error("Error on search",exc_info=True)
			raise
		
		return results



from File import File

class Directory(File):
	"""The built-in type for directories"""
	
	def __init__(self,filename,mimetype,volume):
		File.__init__(self,filename=filename,mimetype=mimetype,volume=volume)

	def __str__(self):
		return "directory %s"%(self.get_filename())



# from threading import RLock
from MetadataService.ProducerConsumer import Producer
from MetadataService.Component import Component
from VolumeExpert import _VolumeExpert
from VolumeCrawler import VolumeCrawler
from InfospaceObjects import Volume

class VolumeCrawlerManager(Component,Producer):

# 	lock = RLock()
	crawler_list = {}
	
	def __init__(self,expert):
		#Contract
		assert isinstance(expert,_VolumeExpert)
		
		Component.__init__(self,public_interface=None) #no interface for this one, dude
		self.set_component_name("FilesystemAgent.VolumeCrawlerManager")

		Producer.__init__(self,5) # FIXME make queue size configurable
		self.expert = expert

	def get_crawler(self,volume):
		assert isinstance(volume,Volume)
# 		self.lock.acquire()
		try: c = self.crawler_list[volume]
		except KeyError: c = self.__start_crawler(volume)
# 		self.lock.release()
		return c
	
	def kill_crawler(self,volume):
		assert isinstance(volume,Volume)
# 		self.lock.acquire()
		self.crawler_list[volume].kill()
		del(self.crawler_list[volume])
# 		self.lock.release()
		
	def kill_all_crawlers(self):
# 		self.lock.acquire()
		crawlers = self.crawler_list.values()
		for a in crawlers:
			self.log.debug("Killing crawler %s",a)
			a.kill()
			self.log.debug("Killed crawler %s",a)
			
		self.log.debug("Killed all crawlers")
		self.crawler_list = {}
# 		self.lock.release()
	
	def __start_crawler(self,volume):
# 		self.lock.acquire()
		self.crawler_list[volume] = VolumeCrawler(volume=volume,
					produce_method=self.produce,expert=self.expert)
					# hack to make the volume crawlers produce through the volume crawler manager
		self.crawler_list[volume].start()
		a = self.crawler_list[volume]
# 		self.lock.release()
		return a
		
	def get_crawlers(self):
# 		self.lock.acquire()
		v = self.crawler_list.values()
# 		self.lock.release()
		return v



##############################################################################
#
# Copyright (c) 2001, 2002 Zope Corporation and Manuel Amador (Rudd-O)
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.0 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE
#
##############################################################################
"""Very Simple directory-based ZODB storage

This storage provides for use of dbm files and pickled files as storages that
don't support versions or Undo.  This may be useful when implementing
objects like hit counters that don't need or want to participate
in undo or versions.  Or my metadata indexing project =)

Instructions: pour in a jar, and add ReiserFS to suit your taste.
"""
__version__='$Revision: 1.4 $'[11:-2]

# FIXME need a way to delete temporary files upon start.  Pack() also needs to detect unlinked files.  Since it will be slow, consider putting in an ultrapack() method.  Perhaps it can verify if each file is mentioned in the _index map
from ZODB.MappingStorage import MappingStorage
from ZODB.BaseStorage import BaseStorage
import os
import logging
import time

class DirectoryStorage(MappingStorage):

	def __init__(self, filename):
		self.debuglog = logging.getLogger("DirectoryStorage")
		MappingStorage.__init__(self, filename)
		self.basedir = filename
		if not os.path.isdir(self.basedir):
			os.mkdir(self.basedir)
		import stat
		os.chmod(self.basedir,stat.S_IRUSR|stat.S_IWUSR|stat.S_IXUSR)
		self._index=DirDict(self.basedir)
		self._tindex=[]
		
		try:
			ofilename = os.path.join(self.basedir,"max-oid")
			ofile = file(ofilename,"r")
			self._oid = ofile.read(-1)
			ofile.close()
			os.unlink(ofilename)
		except:
			logging.getLogger("DirectoryStorage").info("Recovering from a previous crash")
			keys=self._index.keys()
			if keys: self._oid=max(keys)
			logging.getLogger("DirectoryStorage").info("Recovery finished")
			
	def close(self):
		BaseStorage.close(self)
		try:
			ofilename = os.path.join(self.basedir,"max-oid")
			ofile = file(ofilename,"w")
			ofile.write(self._oid)
			ofile.close()
		except:
			raise

	def _finish(self, tid, user, desc, ext):
		thetime = time.time()
		theclock = time.clock()
		oids = [ a[0] for a in self._tindex ]
		self._index.start_transaction(tid,oids)
		MappingStorage._finish(self,tid,user,desc,ext)
		self._index.finish_transaction(tid,oids)
		self.debuglog.debug("Committed %d objects in %s sec (%s cputime).",len(self._tindex),time.time() - thetime,time.clock() - theclock)
		
# this is the persistent, transactional dictionary I am using


import cPickle

class DirDict(dict):
	
	def __init__(self,directory):
		if not os.path.isdir(directory):
			e = os.error()
			e.errno = 2
			e.filename = directory
			e.strerror = "Cannot open directory %s"%repr(directory)
			raise IOError, e
		dict.__init__(self)
		self.directory = directory
		self.transdir = os.path.join(directory,"transactions")
		self.recover_from_crash()
		
	def recover_from_crash(self):
		if not os.path.exists(self.transdir):
			os.mkdir(self.transdir)
			return
		tfiles = os.listdir(self.transdir)
		if len(tfiles) == 0: return # everything is well in the land of the lounge lizards
		tfiles = [ os.path.join(self.transdir,tfile) for tfile in tfiles ]
		# recovery procedure
		for tfile in tfiles:
			logging.getLogger("DirDict").info("Rolling transaction %s back",tfile)
			f = file(tfile,"r")
			keys = cPickle.load(f)
			f.close()
			paths = [self.__map_key_to_pathname(key) for key in keys]
			for pathname in paths:
				try: os.rename(pathname+".backup",pathname)
				except: pass
			for pathname in paths:
				try: os.unlink(pathname+".tmp")
				except: pass
			os.unlink(tfile)
			logging.getLogger("DirDict").info("%s non-atomic changes successfully rolled back",len(keys))
		
	def __map_key_to_filename(self,key):
		return key.encode("hex_codec")

		dumb = [ str(ord(byte)) for byte in key ]
		lastcomponent = dumb.pop()
		rest = "-".join(dumb)
		return os.path.join(lastcomponent,rest)
	
	def __map_filename_to_key(self,filename):
		return filename.decode("hex_codec")

		key = ""
		lastitem,arr = filename.split("/")#fixme unixism
		arr = arr.split("-")
		if len(arr) != 7: raise ValueError, "Invalid filename %s"%filename
		for char in arr + [lastitem]:
			byte = chr(int(char))
			key = key + byte
		return key
	
	def __map_key_to_pathname(self,key):
		return os.path.join(self.directory,self.__map_key_to_filename(key))
	
	def __getitem__(self,key):
		assert len(key) == 8
		pathname = self.__map_key_to_pathname(key)
		try:
			f = file(pathname,"rb",0)
			data = f.read(-1)
			f.close()
			return data
		except IOError,e:
			if e.errno == 2: raise KeyError, key
			else: raise
	
	def update(self,dictionary):
		for key,val in dictionary.iteritems():
			self[key] = val

	def __setitem__(self,key,value):
		assert len(key) == 8
		pathname = self.__map_key_to_pathname(key)
		try: os.rename(pathname,pathname+".backup")
		except: pass
		f = file(pathname+".tmp","wb",0)
		f.write(value)
		f.close()
		
	def start_transaction(self,transid,keys):
		assert type(transid) is str
		assert type(keys) is list
		transid = transid.encode("hex_codec")
		tfile = os.path.join(self.transdir,transid)
		tfile = file(tfile,"w")
		cPickle.dump(keys,tfile)
		tfile.close()
		
	def finish_transaction(self,transid,keys):
		assert type(transid) is str
		transid = transid.encode("hex_codec")
		tfile = os.path.join(self.transdir,transid)
		paths = [self.__map_key_to_pathname(key) for key in keys]
		for pathname in paths: os.rename(pathname+".tmp",pathname)
		for pathname in paths:
			try: os.unlink(pathname+".backup")
			except: pass
		os.unlink(tfile)
	
	def __delitem__(self,key):
		# where the hell is this getting called?
		# should this be in the context of a transaction as well?
		# i think this is called during pack()
		# so it needs to be transactional, so deletions don't fuck up
		# the database if interrupted at some point.
		assert len(key) == 8
		pathname = self.__map_key_to_pathname(key)
		try: os.unlink(pathname)
		except OSError,e:
			if e.errno == 2: raise KeyError, key
			else: raise
	
	def keys(self):
# 		dirs = [ str (a) for a in range(256) ]
# 		files = []
		files = os.listdir(self.directory)
# 		for directory in dirs:
# 			dirfiles = os.listdir(self.directory + "/" + directory) # fixme unixism
# 			files.extend( [ directory + "/" + dirfile for dirfile in dirfiles ] )
		keys = []
		for file in files:
			if len(file) != 16: continue # we perhaps need a more stringent test? FIXME
			try: keys.append(self.__map_filename_to_key(file))
			except ValueError: pass
		return keys
	
	def __nonzero__(self): return True




from Service import Service
from PublicInterface import PublicInterface

class InterfaceCollection:
	def __init__(self,interfaces):
		assert type(interfaces) is list
		assert len(interfaces) > 0
		for i in interfaces: assert isinstance(i,PublicInterface)
		for i in interfaces: setattr(self,i.get_name(),i)
			# which means, for each interface, set an attribute for the interface in this class

class InterfaceServer(Service):
	"""The base class for all interface servers. Interface servers follow a Service pattern."""
	"""All passed interfaces are registered in the self.public_interfaces attribute.  Your particular interface server should map incoming method calls to members of the self.public_interfaces instance."""
	
	def __init__(self,config,interfaces):
		Service.__init__(self) # this does not expose any public interface
		self.public_interfaces = InterfaceCollection(interfaces)
		self.config = config



from VolumeCrawlerManager import VolumeCrawlerManager

from MetadataService.InfospaceAgent import InfospaceAgent
from MetadataService.ProducerConsumer import Consumer, Empty
from FilePluginManager import FilePluginManager
from VolumeExpert import VolumeExpert
from VolumeCrawlerManager import VolumeCrawlerManager
from InfospaceObjects import FilesystemInfospace, Volume
from platform import strip_mountpoint, is_or_is_below
import time
import os
import ConfigParser
# FIXME: wrap this database error and other database errors produced by postgresql in nice Catalog errors, and replace this stuff with the replaced Catalog errors
# in other words, this is just a band-aid
from pgdb import DatabaseError

class FilesystemAgent(InfospaceAgent,Consumer):

	file_plugin_manager = None
	log = None
	crawler_manager = None
	volume_expert = None
	catalog = None

	def __init__(self,index,config):
		InfospaceAgent.__init__(self,public_interface=None,index=index,infospace_name="filesystem",config=config)
		self.set_component_name("FilesystemAgent")
		Consumer.__init__(self)
		self.file_plugin_manager = FilePluginManager()

		try:
			stoplist = self.config.get("filesystem infospace","exclude")
			if stoplist: stoplist = stoplist.split(":")
			else: stoplist = []
		except (ConfigParser.NoOptionError,ConfigParser.NoSectionError):
			stoplist = []
		dbdir = self.config.get("index","path")
		stoplist = stoplist + [ dbdir ]

		self.log.debug("Using stoplist %s",stoplist)
		self.volume_expert = VolumeExpert(stoplist)
		self.crawler_manager = VolumeCrawlerManager(self.volume_expert)
		self.consume_from(self.crawler_manager)
		
		# FIXME make the db name configurable
		# FIXME make the metadata service start its own postgres daemon
		
	def check_index_schema(self):
		root = self.context.root()
		self.infospaces = root["infospaces"]
		if not self.infospaces.has_key(self.infospace_name):
			self.context.begin_transaction()
			self.log.debug("Registering %s infospace",self.infospace_name)
			self.infospaces[self.infospace_name] = FilesystemInfospace()
			self.context.commit_transaction()
		self.infospace = self.infospaces[self.infospace_name]

	def register_currently_known_volumes(self):
		self.volumes = self.infospace.volumes
		known_volumes = self.volume_expert.get_currently_known_volumes()
		for guid,vr in known_volumes.iteritems():
			# we will not register network volumes, ever
			# we will not register loop-mounted volumes for the time being FIXME?
			# we will also not register removable volumes FIXME
			if vr.loop_mounted:
				self.log.debug("Not registering volume %s (loop-mounted)",guid)
				continue
			if vr.network_volume:
				self.log.debug("Not registering volume %s (network volume)",guid)
				continue
			if vr.removable:
				self.log.debug("Not registering volume %s (removable)",guid)
				continue
			if not self.volumes.has_volume(guid):
				self.context.begin_transaction()
				mp = self.volume_expert.get_mountpoint(guid)
				volume = Volume(guid=guid,mountpoint=mp,filesystem_type=vr.filesystem_type,
							label=vr.label,removable=vr.removable)
				self.log.debug("Registering volume %s",volume)
				self.volumes.store_volume(volume)
				self.context.commit_transaction()
			else:
				self.context.begin_transaction()
				volume = self.volumes.get_volume(guid)
				mp = self.volume_expert.get_mountpoint(guid)
				volume.update_mountpoint(mp)
				self.context.commit_transaction()

	def start_crawlers(self):
		for volume in self.volumes.itervalues():
			if volume.is_mounted():
				if self.volume_expert.is_appropriate(volume.get_mountpoint()):
					if volume.fully_indexed is False:
						self.log.debug("Starting crawler on %s",volume)
						self.crawler_manager.get_crawler(volume)
					else: self.log.debug("Skipping crawler on %s (fully indexed)",volume)
				else: self.log.debug("Skipping crawler on %s (inappropriate)",volume)
			else: self.log.debug("Skipping crawler on %s (unmounted)",volume)

	def kill_crawlers(self):
		self.crawler_manager.kill_all_crawlers()
		
	def run(self):
		if 0:
			import hotshot
			prof = hotshot.Profile(self.get_component_name()+'.prof')
			retval = prof.runctx('self._run()', globals(),locals())
			prof.close()
		else: self._run()

	def _run(self):
		self.context = self.index.acquire_context()
		self.context.set_catalog_infospace("filesystem")
		
		self.check_index_schema()
		self.register_currently_known_volumes()
		
		# store volumes FIXME should also update them
		
		self.start_crawlers()
		
		while not self.killed():
			import time
			times = time.time()
			try:
				# should we follow the "proper" or the dirty way?
				# we could make it so the crawlers send an (0,"volumename") tuple when they are done indexing files, and here, if type(filename[0]) is int, the volume is done, which addiontially means cuz of the semantics of this shit, that the volume can be considered fully crawled, and the next time there's no need to start a crawler for that volume.   tal vez haya mejor IPC methods que esto!  Perhaps we should send a CrawlFinishedEvent object (se me ocurre ahorita) con un puntero al crawler en cuestion.
				# we could also make FileDiscovered, FileDeleted, FileUpdated, VolumeMounted, VolumeUnmounted events
				# The crawler manager sends FileDiscovered and FilesystemFullyCrawled events
				event = self.consume(1)  # this one sends FileDiscovered and FilesystemFullyCrawled events
# 				self.log.debug("Lingered %s sec. before consuming",time.time() - times)
				# I think we should try to act upon volume notifications first, then file notifications and then crawler notifications, in that order, to preemptively act upon volume unmounts, and to keep the index fresh with recent files.
				# we should also extend this mechanism to the file_update_monitor and volume_monitor in the future.
				if event[0] == "FileFound":
					self.index_file(event[1],event[2])
				elif event[0] == "FullyIndexed":
					self.mark_volume_as_fully_indexed(event[1])
				else: raise Exception, "Unknown event %s" % event
# 				self.log.debug("Cycle completed in %s sec.",time.time()-times)
				self.sleep()
			except Empty:
				self.log.debug("No files to crawl, lingering")
			except:
				self.log.error("Anomalous exception while indexing, aborting filesystem infospace agent",exc_info=True)
				break
			
		
		self.kill_crawlers()
		self.context.release()
		del (self.context)
		
	def mark_volume_as_fully_indexed(self,volume):
		self.log.debug("Volume %s is now fully indexed",volume)
		self.context.begin_transaction()
		volume.fully_indexed = True
		self.context.commit_transaction()
		self.log.debug("Volume %s indexed? %s",volume,repr(volume.fully_indexed))

	def index_file(self,filename,volume):
		self.log.debug("Starting with %s", filename)
		thetime = time.time()
		key = strip_mountpoint(filename,volume.get_mountpoint())
		cat_key = "|".join([volume.guid,key])
		file = None
		try:
			if volume.has_file(key) and not volume.has_file_changed_since(key,os.stat(filename).st_mtime):
				self.log.debug("%s up-to-date, skipped in %s sec.", filename,time.time() - thetime)
				return
			file = self.file_plugin_manager.get_information_object(filename=filename,volume=volume)

		except IOError,e:
			if e.errno == 2:
				self.log.debug("Removing %s from database (IOError)",filename,exc_info=True)
				if volume.has_file(key):
					self.context.begin_transaction()
					self.context.uncatalog_object(cat_key)
					volume.remove_file(key)
					self.context.commit_transaction()
					self.log.debug("Removed %s from database (no longer exists) in %s sec.",filename,time.time() - thetime)
			elif e.errno == 13: self.log.debug("Permission to open %s denied ",filename)
			else: self.log.warning("Metadata extraction failed for %s",filename,exc_info=True)
 		except:
 			self.log.warning("Metadata extraction failed for %s",filename,exc_info=True)

		if not file:
			 self.log.debug("Skipping %s (decided in %s sec.)", filename,time.time() - thetime)
			 return # because all cases have been handled before this
		
		# if the file was actually gotten
		self.context.begin_transaction()
		try:
			volume.store_file(file)
			self.context.catalog_object(cat_key,file)
		except DatabaseError:
			self.log.warning("Cannot store %s in the catalog",file,exc_info=True)
			self.context.abort_transaction()
		except:
			self.context.abort_transaction()
			raise
		
		self.context.commit_transaction()
		self.log.debug("Indexed %s in %s sec." , file , time.time() - thetime)
	
	def search(self,expression,context):
		from pgtools.query import Expression, optimize
		from MetadataService.Index import IndexContext
		from sets import Set
		assert isinstance(expression,Expression)
		assert isinstance(context,IndexContext)
		
		thetime = time.time()
		self.log.debug("Asked to search for %s",expression)
		
		volumes = context.root()["infospaces"]["filesystem"].volumes
		
		expression = decompose_into_volumes(expression,volumes.values())
		
		self.log.debug("Final expression is %s (computed in %s sec.)",expression,time.time()-thetime)
		
		
		thetime = time.time()
		
		context.set_catalog_infospace("filesystem")
		results = context.search(expression)
		assert isinstance(results,Set)
		
		self.log.debug("Catalog returned %s results in %s sec.",len(results),time.time() -thetime)
		
		thetime = time.time()
		resultset = []
		
		from urllib import pathname2url
		mpcache = {}
		for guid,volume in volumes.iteritems():
			mountpoint = volume.get_mountpoint()
			if mountpoint: mpcache[guid] = pathname2url(mountpoint)
		
		
		for result in results:
		# convert each result to an URL, which is a string encoded in platform encoding
			vguid,fpath = result.split("|",1)
# 			fpath = pathname2url(fpath)
			if mpcache[vguid]: path = "file://" + os.path.join(mpcache[vguid],fpath)
			else: path = "volume://"+vguid+"/"+fpath
			if path.endswith("/") and path != "file:///": path = path[:-1]
			resultset.append(path)
		self.log.debug("Total number of results: %s (paths generated in %s sec.)",len(resultset),time.time() -thetime)
		
# 		if len(resultset) < 10:
# 			for a in resultset: print "Result:", a # just for debugging!
		
		return resultset


from pgtools.query import Aggregation,Criterion,Equals,IsBelow,IsOrIsBelow,Or
def decompose_into_volumes(expression,volumes):

	if isinstance(expression,Aggregation):
		newexpressions = []
		for e in expression.expressions:
			exp = decompose_into_volumes(e,volumes)
			if exp is not None: newexpressions.append(exp)
		if len(newexpressions): expression.expressions = newexpressions # replace with the new expressions
		else: raise ValueError, "Expression %s validated to an empty expression" # invalid query by the user
		return expression
	
	elif isinstance(expression,Criterion):
		if expression.attribute != "filename": return expression
		filename = os.path.normpath(expression.value)
		
		if isinstance(expression.operator,Equals):
			volume = deepest_volume_for(filename,volumes)

			if volume is None: raise ValueError,"%s is not an absolute path"%filename
			mountpoint = volume.get_mountpoint()
			if mountpoint is None: raise RuntimeError, "%s got unmounted in the middle of an operation"%volume
			
			relative_path = strip_mountpoint(filename,mountpoint)
			
			
			
			exp1 = Criterion()
			exp1.attribute = "absolute_path"
			exp1.operator = Equals()
			exp1.value = os.path.normpath("/%s/%s"%(volume.guid,relative_path))
			
			return exp1
			
		elif isinstance(expression.operator,IsBelow) or isinstance(expression.operator,IsOrIsBelow):
		
			involved_volumes = volumes_below(filename,volumes)
			if involved_volumes is None: raise ValueError,"%s is not an absolute path"%filename
			
			
			volume_exps = []
			
			shortest_mountpoint = None
			for iv in involved_volumes: 
				if shortest_mountpoint is None:
					shortest_mountpoint = iv.get_mountpoint()
					if shortest_mountpoint is None: raise RuntimeError, "%s got unmounted in the middle of an operation"%iv
				else:
					if len(iv.get_mountpoint()) < len(shortest_mountpoint): shortest_mountpoint = iv.get_mountpoint()
					
			for iv in involved_volumes:
				mountpoint = iv.get_mountpoint()
				if mountpoint is None: raise RuntimeError, "%s got unmounted in the middle of an operation"%iv
				relative_path = strip_mountpoint(filename,mountpoint)
				
				exp2 = Criterion()
				exp2.attribute = "absolute_path"
				if isinstance(expression.operator,IsBelow) and mountpoint != shortest_mountpoint:
					exp2.operator = IsOrIsBelow()
				else: exp2.operator = expression.operator
				exp2.value = os.path.normpath("/%s/%s"%(iv.guid,relative_path))
					
				volume_exps.append(exp2)
			
			return Or(volume_exps)
		
		else: raise NotImplementedError, "operator %s not implemented yet for filename queries"%expression.operator
		
	else: raise RuntimeError, "What the fuck?"


# consider moviing this piece of shit to the platform modules
def deepest_volume_for(filename,volumes):
	# this may return None at some point!!!!
	filename = os.path.normpath(filename)
	mounted_volumes = [ v for v in volumes if v.get_mountpoint() is not None ]
	matching_volumes = [ v for v in mounted_volumes if is_or_is_below(filename,v.get_mountpoint()) ]
	if len(matching_volumes) == 0: return None
	deepest_volume = matching_volumes[0]
	for v in matching_volumes:
		if len(v.get_mountpoint()) > len(deepest_volume.get_mountpoint()): deepest_volume = v
	return deepest_volume
	
def volumes_below(path,volumes):
	# this may return None at some point!!!!
	path = os.path.normpath(path)
	mounted_volumes = [ v for v in volumes if v.get_mountpoint() is not None ]
	matching_volumes = [ v for v in mounted_volumes if is_or_is_below(v.get_mountpoint(),path)  ]
	if len(matching_volumes) == 0:
		deepest_volume = deepest_volume_for(path,volumes)
		if deepest_volume is None: matching_volumes = None
		else: matching_volumes = [ deepest_volume_for(path,volumes) ]
	return matching_volumes



############################################################################
#    Copyright (C) 2005 by Manuel Amador (Rudd-O)                          #
#    rudd-o@amautacorp.com                                                 #
#                                                                          #
#    This program is free software; you can redistribute it and#or modify  #
#    it under the terms of the GNU General Public License as published by  #
#    the Free Software Foundation; either version 2 of the License, or     #
#    (at your option) any later version.                                   #
#                                                                          #
#    This program is distributed in the hope that it will be useful,       #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of        #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         #
#    GNU General Public License for more details.                          #
#                                                                          #
#    You should have received a copy of the GNU General Public License     #
#    along with this program; if not, write to the                         #
#    Free Software Foundation, Inc.,                                       #
#    59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             #
############################################################################

# FIXME implement an interface that allows the indexer to compare in-store file objects' interface versions with its corresponding plugin"s interface version, to provide for automatic schema migration.  Bump the interface number on your plugin, and pof all files are automatically reindexed when they are modified (or perhaps upon startup/reload)
# FIXME implement an observer pattern so that the index database detects when the plugins are reloaded, suspend database access to client classes, run __prepare_indexes and refresh objects whose interfaces have changed (ay que pensar bien el orden de mierda de estas operaciones para que no se vaya todo a la verga"), 

import sys
import os
import glob
import imp
import commands
from MetadataService import common

from MetadataService.InfospaceAgents.Filesystem.platform import is_executable

if not is_executable("extract2"): raise ImportError, "libextractor is not available"

def get_metadata(filename):
		""" Returns a dictionary with the metadata keys and values for the file.  All strings are platform-encoded. """
		command = "extract %s"%common.shell_quote(filename)
		output = commands.getoutput(command)
		output = output.splitlines()
		returned = {}
		for line in output:
			try: key,value = line.split("-",2)
			except: continue
			returned[key.strip()] = value.strip()
		return returned

# fixme subpar implementation
def get_mimetype(filename):
		metadata = get_metadata(filename)
		if metadata.has_key("mimetype"): return metadata["mimetype"]


############################################################################
#    Copyright (C) 2005 by Manuel Amador (Rudd-O)                          #
#    rudd-o@amautacorp.com                                                 #
#                                                                          #
#    This program is free software; you can redistribute it and#or modify  #
#    it under the terms of the GNU General Public License as published by  #
#    the Free Software Foundation; either version 2 of the License, or     #
#    (at your option) any later version.                                   #
#                                                                          #
#    This program is distributed in the hope that it will be useful,       #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of        #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         #
#    GNU General Public License for more details.                          #
#                                                                          #
#    You should have received a copy of the GNU General Public License     #
#    along with this program; if not, write to the                         #
#    Free Software Foundation, Inc.,                                       #
#    59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             #
############################################################################

# FIXME implement an interface that allows the indexer to compare in-store file objects' interface versions with its corresponding plugin"s interface version, to provide for automatic schema migration.  Bump the interface number on your plugin, and pof all files are automatically reindexed when they are modified (or perhaps upon startup/reload)
# FIXME implement an observer pattern so that the index database detects when the plugins are reloaded, suspend database access to client classes, run __prepare_indexes and refresh objects whose interfaces have changed (ay que pensar bien el orden de mierda de estas operaciones para que no se vaya todo a la verga"), 

import sys
import os
import glob
import imp

from MetadataService.common import utc_timestamp_to_local,convert_to_unicode,convert_from_unicode
import locale

# FIXME PRETTY PLEASE
# this may involve the creation of two other data types: naivedate (yy/mm/dd) and naivetime(hh:mm:ss) which operate in local time, and then client methods can interpret them and cast them to DateTime on their own, with this knowledge.
# or perhaps even better create an object daydelta = num of seconds after 0h00 and then we can sum that to the DateTime
# this should be moved to KDE's own module
import qt

def convert_qvariant(qvar,key=None):
	if qvar.type() == 3:
		# Guaranteed that the returned string is a string encoded in the platform's encoding
		return convert_qstring_to_string(qvar.toString())
	elif qvar.type() == 16:
		return qvar.toInt()
	elif qvar.type() == 18:
		return qvar.toBool()
	elif qvar.type() == 34:
		return qvar.toDouble()
	elif qvar.type() == 26:
		d = qvar.toDate()
		return DateTime.DateTime(d.year(),d.month(),d.day(),0,0,0) # QDate assumes local timezone
	elif qvar.type() == 27:
		return None # FIXME according to above instructions
	else:
		if key: key = " for key %s" % key
		raise ValueError, "QVariant unknown value " + str(qvar.type()) + key

def convert_qstring_to_string(qstring):
	""" Returns a regular string from a QString.  The returned string is encoded in the platform's preferred encoding """
	if not isinstance(qstring,qt.QString): raise ValueError
	return convert_from_unicode(unicode(qstring),locale.getpreferredencoding())

def convert_qstring_to_utf8(qstring):
	""" Returns a regular string from a QString.  The returned string is encoded in UTF-8  """
	if not isinstance(qstring,qt.QString): raise ValueError
	return convert_from_unicode(unicode(qstring),"UTF-8")

def convert_qstring_to_unicode(qstring):
	""" Returns a regular string from a QString.  The returned string is encoded in the platform's preferred encoding """
	if not isinstance(qstring,qt.QString): raise ValueError
	return unicode(qstring)


# this should make possible that plug-ins that need the KDE expert do not get imported
import kdecore
import kio
import threading
#from qt import QTimer
kdecore.KCmdLineArgs.init(["metadata-service"],"metadata-service","Metadata service","0.1",True)
k_app = kdecore.KApplication(False,False)
#k_timer = QTimer(None)
k_app_thread = threading.Thread(target=k_app.exec_loop)
k_app_thread.setDaemon(True)
k_app_thread.start()
#k_timer.start(200,False)

def get_mimetype(filename):
		""" Returns a file's MIME type. """
		f = convert_to_unicode(filename)
		kmimetype = kio.KMimeType.findByPath(f)
		mimetype = convert_qvariant(kmimetype.property("Name"))
		return mimetype

def get_metadata(filename):
		""" Returns a dictionary with the metadata keys and values for the file.  All strings are platform-encoded. """
		f = convert_to_unicode(filename)
		k = kio.KFileMetaInfo(f)
		returned = {}
		for groupname in k.groups():
			group = k.group(groupname)
			for keyname in group.keys():
				#try:
					kkeyname = "%s.%s" % (str(groupname),str(keyname))
					kkeyname = kkeyname.lower()
					value = convert_qvariant(k.value(keyname),keyname)
					returned[kkeyname] = value
				#except ValueError,e:
				#	print "WARNING: %s on key %s" % (str(e),keyname)
		return returned

def get_mounted_mountpoints(): #FIXME shouldn't this be converted to a plain list of strings before returning?????
	return [ convert_qstring_to_string(a.mountPoint()) for a in kdecore.KMountPoint.currentMountPoints(1) ]



#!/usr/bin/env python

from AdaptiveLoadController import AdaptiveLoadController

import time
class AdaptiveSleeper:
	"""A mixin class you can inherit in your thread-derived classes. Whenever you want your thread to sleep an optimum time to maintain the target load specified in the AdaptiveLoadController, call self.sleep()."""
	
	def __init__(self):
		pass

	def sleep(self):
		self.__do_sleep()

	def __do_sleep(self):
		has_slept_for = 0.0
		while has_slept_for < AdaptiveLoadController.throttle_ms:
			difference = AdaptiveLoadController.throttle_ms - has_slept_for
			if difference <= 0.0: break
			now_sleep_for = min(AdaptiveLoadController.sleeper_granularity,difference)
			time.sleep(float(now_sleep_for))
			has_slept_for = has_slept_for + now_sleep_for




import ConfigParser
import os
import sys
import os
import logging
from logging.handlers import SysLogHandler as sl

def get_prefix(): return os.path.abspath(os.path.join(os.path.dirname(sys.argv[0]),".."))

def get_data_dir():
	prefix = get_prefix()
	if prefix == "/usr": return "/var"
	return os.path.join(prefix,"var")

def get_sysconfdir():
	prefix = get_prefix()
	if prefix == "/usr": return "/etc"
	return os.path.join(prefix,"etc")

config_filename = "metadata-service.conf"

def get_configfile():
	paths = [get_sysconfdir(),"/usr/local/etc","/etc"]
	for path in paths:
		sysconffile = os.path.join(path,config_filename)
		if os.path.exists(sysconffile):
			return sysconffile

class MetadataServiceConfig(ConfigParser.ConfigParser):
 
	config_filename = None
	data_dir = None

	def get_filename(self):
		return self.config_filename

	def get_data_dir(self):
		return self.data_dir

	def get_socket(self):
		# FIXME: make the data dir not relative to the binary (which may be lockated in different places) but relative to this file itself.
		try: sockaddr = self.get("socket interface","socket")
		except (ConfigParser.NoSectionError,ConfigParser.NoOptionError):
			sockaddr = os.path.join(self.get_data_dir(),"socket")
		return sockaddr

 	def __init__(self,filename=None):
 		ConfigParser.ConfigParser.__init__(self)
		if filename is not None:
			self.config_filename = os.path.abspath(filename)
		else:
			self.config_filename = get_configfile()
		if self.config_filename is not None: self.read(self.config_filename)
		self.data_dir = os.path.join(get_data_dir(),"lib","metadata-service")
 		self.apply_defaults()
 
 	def apply_defaults(self):
		if not self.has_section("index"): self.add_section("index")
 		try: dbdir = self.get("index","path")
 		except ConfigParser.NoOptionError:
 			dbdir = os.path.join(self.get_data_dir(),"index")
			self.set("index","path",dbdir)

		if not self.has_section("logging"): self.add_section("logging")
 		try:
			level = self.get("logging","log level")
			l = { "INFO":logging.INFO,
				"DEBUG":logging.DEBUG,
				"ERROR":logging.ERROR,
				"CRITICAL":logging.CRITICAL,
				"WARNING":logging.WARNING
			    }
			self.set("logging","log level",str(l[level]))
 		except ConfigParser.NoOptionError: self.set("logging","log level",str(logging.INFO))
		except KeyError: raise ValueError, "Invalid log level %s"%level

		try: 
			syslog = self.getboolean("logging","use syslog")
		except ConfigParser.NoOptionError: 
			self.set("logging","use syslog","True")

 		try:
			facility = self.get("logging","log facility")
			facilities = {'ALERT':sl.LOG_ALERT,
				'AUTH':sl.LOG_AUTH, 'AUTHPRIV':sl.LOG_AUTHPRIV,
				'CRON':sl.LOG_CRON, 'DAEMON':sl.LOG_DAEMON,
				'KERN':sl.LOG_KERN, 'LOCAL0':sl.LOG_LOCAL0,
				'LOCAL1':sl.LOG_LOCAL1, 'LOCAL2':sl.LOG_LOCAL2,
				'LOCAL3':sl.LOG_LOCAL3, 'LOCAL4':sl.LOG_LOCAL4,
				'LOCAL5':sl.LOG_LOCAL5, 'LOCAL6':sl.LOG_LOCAL6,
				'LOCAL7':sl.LOG_LOCAL7, 'LPR':sl.LOG_LPR,
				'MAIL':sl.LOG_MAIL, 'NEWS':sl.LOG_NEWS,
				'SYSLOG':sl.LOG_SYSLOG, 'USER':sl.LOG_USER,
				'UUCP':sl.LOG_UUCP
			}
			self.set("logging","log facility",str(facilities[facility]))
 		except ConfigParser.NoOptionError: self.set("logging","log facility",str(sl.LOG_DAEMON))
		except KeyError: raise ValueError, "Invalid log facility %s"%facility

		if not self.has_section("load controller"): self.add_section("load controller")
 		try:
			try:
				load = self.getfloat("load controller","target load")
			except ValueError:
				load = self.get("load controller","target load")
				raise ValueError, "Invalid value %s for target load"%load
			if load < 0.1 or load > 10:
				raise ValueError, "Invalid value %s for target load"%load
 		except ConfigParser.NoOptionError: self.set("load controller","target load",str(1.0))





def get(config_filename=None):
	return MetadataServiceConfig(config_filename)



from MetadataService.common import DateTime # FIX times/dates everywhere
from MetadataService.Database import Database

import ZODB
from ZODB import DB
# from ZODB.FileStorage import FileStorage
# from BDBStorage.BDBMinimalStorage import BDBMinimalStorage
# from BDBStorage.BerkeleyBase import BerkeleyConfig
from DirectoryStorage import DirectoryStorage
from MetadataService.InfospaceList import InfospaceList
import os

# FIXME implement referential integrity
class ZODBDatabase(Database,DB):

	def __init__(self,path):
		self.__storage = DirectoryStorage(path)
# 		config = BerkeleyConfig()
# 		config.cachesize = 128*1024
# 		self.__storage = BDBMinimalStorage("MetadataServiceStorage",path,config=config)
		DB.__init__(self,self.__storage)
	
	def open_connection(self):
		return self.open()
		
	def root(self,connection):
		return connection.root()
		
	def close_connection(self,connection):
		connection.close()

	def update_schema(self):
		conn= self.open()
		root = conn.root()
		if not root.has_key("infospaces"):
			self.begin_transaction(conn)
			root["infospaces"] = InfospaceList()
			self.commit_transaction(conn)
		conn.close()
		
	def begin_transaction(self,connection):
		a = get_transaction()

	def abort_transaction(self,connection):
		get_transaction().abort()

	def commit_transaction(self,connection):
		get_transaction().commit()

	def close(self):
		DB.close(self)
		
	def pack(self):
		DB.pack(self)


# data model used here
# 
# map key to object (frozen,serialized object) to be able to unserialize it later
# "/home/music/A-Ha/Take on me.mp3", "lsajfls adjflkasdjfa sfjklasjf oiwejflfj pickled data"
# 
# map key,property,value
# 
# "/home/music","artist","something"
# 
# several columns, all may be NULL, depending on data type:
# value_text  (TEXT)
# value_datetime   (DATETIME)
# value_int    (INTEGER)
# value_float    (DOUBLE)
# 
# the query processor should be smart enough to transform everything into the proper column and operator
# 
# { "property": "artist" , "value" : "*amagador*" , "operator" : "glob" }
#     should be transformed into
# (name = "artist" and value_text like "%amagador%")
# 
# 
# 
# 
# 



PluginInterface.require_command("pdftotext")
import commands

class PDFDocument(PluginInterface.File):

	contents = None
	
	PDFDocument_catalogable = [ "contents" ]

	def __init__(self,filename,mimetype,volume):
		PluginInterface.File.__init__(self,filename,mimetype,volume)
		# now we'll read the file contents
		command = "pdftotext -enc UTF-8 -raw %s - 2> /dev/null" % PluginInterface.shell_quote(filename)
		status, output = commands.getstatusoutput(command)
		if status == 0 :
			try: self.contents = PluginInterface.utf8_to_platform(output)
			except (UnicodeDecodeError, UnicodeEncodeError):
				PluginInterface.log("pdftotext-application-pdf").error("Error while converting %s to platform encoding",filename,exc_info=True)
		else:
			PluginInterface.log("pdftotext-application-pdf").warning("pdftotext failed on filename %s with status %d",filename, status)
		
	def __str__(self):
		return "PDF document %s"%(self.get_filename())

def get_factory():
	return PDFDocument

def get_mimetypes():
	return ["application/pdf"]



from MetadataService.Service import Service
import os
from VolumeExpert import _VolumeExpert
from MetadataService.AdaptiveSleeper import AdaptiveSleeper
from MetadataService.ProducerConsumer import Full
from InfospaceObjects import Volume
import time

class VolumeCrawler(Service,AdaptiveSleeper):

	volume = None
	indexer = None
	log = None
	files_to_index = None
# 	dev_node = None

	def __init__(self,volume,expert,produce_method):
		# Contract:
		assert isinstance(volume,Volume)
		assert isinstance(expert,_VolumeExpert)
		assert callable(produce_method)
		
		Service.__init__(self,public_interface = None)
		self.set_component_name("FilesystemAgent.VolumeCrawler."+volume.get_mountpoint().replace(".","_"))
		AdaptiveSleeper.__init__(self)
		self.log.debug("Initializing")
# 		self.dev_node = os.stat(volume.get_mountpoint()).st_dev
		self.volume = volume
		self.expert = expert
		self.produce = produce_method # for the moment this is hardcoded, but the ideal sutiation would be that this is an instance of a Producer and the volume crawler manager is a Multi-consumer or some stuff like that
		self.log.debug("Initialized")

	# FIXME signaling the indexer to delete a file is the responsiblity of the process/service that found the file did not exist
	def run(self):
		self.log.info("Started")
		expert = self.expert
		# for the root
		self.add_to_queue(("FileFound",self.volume.get_mountpoint(),self.volume))
		self.sleep()
		# and for the children
		for root, dirs, files in os.walk(self.volume.get_mountpoint()):  
			if self.killed(): break

			for name in dirs[:]:
				if self.killed(): break
				filename = os.path.join(root,name)
				if os.path.ismount(filename): # do not cross FS (volume) boundaries
					self.log.debug("Removing %s from walk list (mountpoint of other volume)",filename)
					dirs.remove(name)
				elif not expert.is_appropriate(filename):
					self.log.debug("Removing %s from walk list (inappropriate)",filename)
					dirs.remove(name)
				
			for name in files:
				if self.killed(): break
				filename = os.path.join(root,name)
				if not expert.is_appropriate(filename):
					self.log.debug("Skipping %s (inappropriate)",filename)
					continue
				self.add_to_queue(("FileFound",filename,self.volume))
				self.sleep()
			
			for name in dirs:
				if self.killed(): break
				filename = os.path.join(root,name)
				self.add_to_queue(("FileFound",filename,self.volume))
				self.sleep()
		
		# at this point, the volume has been fully crawled, but files may still be in the queue for the filesystem crawler to consume.  The ideal case would be that the volume crawler now persists a datum that tells it "Do not crawl this volume on the next startup"
		self.add_to_queue(("FullyIndexed",self.volume))
		
		while not self.killed():
			self.log.debug("Fully crawled, now lingering")
			time.sleep(1)
		self.log.info("Finished")

	def add_to_queue(self,event):
		assert type(event) is tuple
		while not self.killed():
			try:
				self.produce(event,1)
				return
			except Full: pass

